from PymoNNto.NetworkCore.Base import *

import random
#def str_eval_conn(eval_string, synapses, s_id, sx, sy, sz, d_id, dx, dy, dz):
#    return eval(eval_string)

class SynapseGroup(NetworkObjectBase):

    def __init__(self, src, dst, net, connectivity=None, tag=None, partition=False, partition_blocks='auto', behaviour = {}):
        super().__init__(tag=tag)

        if len(src.tags)>0 and len(dst.tags)>0:
            self.add_tag(src.tags[0]+' => '+dst.tags[0])

        if net is not None:
            net.SynapseGroups.append(self)
            self.network = net

        self.recording = True

        self.src = src
        self.dst = dst
        self.enabled = True
        self.group_weighting = 1

        self.description = {'connectivity': str(connectivity), 'partition':str(partition)}

        if connectivity is not None:
            self.set_connectivity(connectivity)

        if partition:
            self.partition(split_size=partition_blocks)

        self.behaviour = behaviour

        for k in self.behaviour:
            if self.behaviour[k].run_on_init:
                self.behaviour[k].set_variables(self)

    def find_objects(self, key):
        result = []

        if key in self.behaviour:
            result.append(self.behaviour)

        for bk in self.behaviour:
            behaviour = self.behaviour[bk]
            result += behaviour[key]

        return result

    def set_connectivity(self, connectivity):
        src = self.src
        dst = self.dst
        self.enabled = self.get_synapse_mat()

        s_id = np.tile(np.arange(src.size), (1, dst.size)).reshape(dst.size, src.size)

        if hasattr(src, 'x') and hasattr(src, 'y') and hasattr(src, 'z'):
            sx = np.tile(src.x, (1, dst.size)).reshape(dst.size, src.size)
            sy = np.tile(src.y, (1, dst.size)).reshape(dst.size, src.size)
            sz = np.tile(src.z, (1, dst.size)).reshape(dst.size, src.size)

        d_id = np.tile(np.arange(dst.size), (1, src.size)).reshape(src.size, dst.size).transpose()

        if hasattr(dst, 'x') and hasattr(dst, 'y') and hasattr(dst, 'z'):
            dx = np.tile(dst.x, (1, src.size)).reshape(src.size, dst.size).transpose()
            dy = np.tile(dst.y, (1, src.size)).reshape(src.size, dst.size).transpose()
            dz = np.tile(dst.z, (1, src.size)).reshape(src.size, dst.size).transpose()

        if hasattr(src, 'width') and hasattr(dst, 'width'):  # scaling
            sxs = dst.width / src.width
            dxs = src.width / dst.width

        if hasattr(src, 'height') and hasattr(dst, 'height'):  # scaling
            sys = dst.height / src.height
            dys = src.height / dst.height

        def in_box(dist):
            x_diff=np.abs(sx-dx)
            y_diff=np.abs(sy-dy)
            z_diff=np.abs(sz-dz)
            return (x_diff<=dist)*(y_diff<=dist)*(z_diff<=dist)

        def in_circle(dist):
            x_diff=np.abs(sx-dx)
            y_diff=np.abs(sy-dy)
            z_diff=np.abs(sz-dz)
            return np.sqrt(x_diff*x_diff+y_diff*y_diff+z_diff*z_diff)<=dist

        if type(connectivity) == str:
            self.enabled = eval(connectivity)
            #print(self.enabled)
        elif type(connectivity) == float or type(connectivity) == int:
            if connectivity >= 0:
                self.enabled = (np.sqrt(
                    np.power(np.abs(sx - dx), 2) + np.power(np.abs(sy - dy), 2)) <= connectivity) * (s_id != d_id)
        else:
            self.enabled = connectivity(self, s_id, sx, sy, sz, d_id, dx, dy, dz)


    def set_var(self, key, value):
        setattr(self,key,value)
        return self

    #def add_tag(self, tag):
    #    self.tags.append(tag)
    #    return self


    def get_synapse_mat_dim(self):
        return self.dst.size, self.src.size#self.get_dest_size(), self.get_src_size()


    def get_synapse_mat(self):
        return np.zeros(self.get_synapse_mat_dim())

    def get_random_synapse_mat_fixed(self, min_number_of_synapses=0):
        dim = self.get_synapse_mat_dim()
        result = np.zeros(dim)
        if min_number_of_synapses != 0:
            for i in range(dim[0]):
                synapses = np.random.choice(list(range(dim[1])), size=int(min_number_of_synapses), replace=False)
                result[i, synapses] = np.random.rand(len(synapses))
        return result#*np.random.rand(dim)

    def get_random_synapse_mat(self, density=1.0, clone_along_first_axis=False, rnd_code=None):
        return self.get_random_nparray((self.get_synapse_mat_dim()), density, clone_along_first_axis, rnd_code=rnd_code)*self.enabled

    def get_synapse_group_size_factor(self, synapse_group, synapse_type):

        total_weighting = 0
        for s in synapse_group.dst.afferent_synapses[synapse_type]:
            total_weighting += s.group_weighting

        total = 0
        for s in synapse_group.dst.afferent_synapses[synapse_type]:
            total += s.src.size*s.src.group_weighting

        return total_weighting/total*synapse_group.src.size*synapse_group.group_weighting

    def get_distance_mat(self, radius, src_x=None, src_y=None, dst_x=None, dst_y=None):
        if src_x is None: src_x = self.src.x
        if src_y is None: src_y = self.src.y
        if dst_x is None: dst_x = self.dst.x
        if dst_y is None: dst_y = self.dst.y

        result_syn_mat = np.zeros((len(dst_x), len(src_x)))

        for d_n in range(len(dst_x)):
            dx = np.abs(src_x - dst_x[d_n])
            dy = np.abs(src_y - dst_y[d_n])

            dist = np.sqrt(dx * dx + dy * dy)
            inv_dist = np.clip(radius - dist, 0.0, None)
            inv_dist /= np.max(inv_dist)

            result_syn_mat[d_n] = inv_dist

        return result_syn_mat


    def get_ring_mat(self, radius, inner_exp, src_x=None, src_y=None, dst_x=None, dst_y=None):
        dm = self.get_distance_mat(radius, src_x, src_y, dst_x, dst_y)
        ring = np.clip(dm - np.power(dm, inner_exp)*1.5, 0.0, None)
        return ring/np.max(ring)

    def get_max_receptive_field_size(self):
        max_dx = 1
        max_dy = 1
        max_dz = 1

        #print(self.dst.x[89])

        #print(self.dst.size, self.dst.x.shape, self.dst.y.shape, self.dst.z.shape)
        #print(self.src.size, self.src.x.shape, self.src.y.shape, self.src.z.shape)

        for i in range(self.dst.size):
            if type(self.enabled) is np.ndarray:
                mask = self.enabled[i]
            else:
                mask = self.enabled

            if np.sum(mask)>0:
                x = self.dst.x[i]
                y = self.dst.y[i]
                z = self.dst.z[i]

                sx_v = self.src.x[mask]
                sy_v = self.src.y[mask]
                sz_v = self.src.z[mask]

                max_dx = np.maximum(np.max(np.abs(x-sx_v)), max_dx)
                max_dy = np.maximum(np.max(np.abs(y-sy_v)), max_dy)
                max_dz = np.maximum(np.max(np.abs(z-sz_v)), max_dz)

        return max_dx, max_dy, max_dz


    def partition(self, split_size='auto'):#, receptive_field_size='auto'

        #if receptive_field_size == 'auto':
        #    receptive_field_size = self.get_max_receptive_field_size()

        if split_size == 'auto':
            best_block_size = 7
            w = int((self.src.width/best_block_size+self.dst.width/best_block_size)/2)
            h = int((self.src.height/best_block_size+self.dst.height/best_block_size)/2)
            d = int((self.src.depth / best_block_size + self.dst.depth / best_block_size) / 2)
            split_size = [np.maximum(w, 1), np.maximum(h, 1), np.maximum(d, 1)]
            if split_size[0]<3 and split_size[1]<3 and split_size[2]<3:
                return

        #print('partition:', receptive_field_size, split_size)

        self.network.partition_Synapse_Group3(self, steps=split_size)

        #self.network.partition_Synapse_Group(self, receptive_field_size=self.get_max_receptive_field_size(), split_size=split_size)


from PymoNNto.NetworkBehaviour.Input.Activator import *
from PymoNNto.NetworkBehaviour.Logic.Basics.Normalization import *

#import matplotlib.pyplot as plt

def identity(neurons):
    return neurons.activity-neurons.TH

def analog_activation_function(neurons):
    return np.clip(np.power(neurons.activity, 2.0), 0, 1)

def binary_activation_function(neurons):
    if neurons.rnd_act_factor != None:
        sig = neurons.rnd_act_factor * np.random.rand(neurons.size)
    else:
        sig = 0

    if hasattr(neurons, 'nox'):
        nox = neurons.nox
    else:
        nox = 0

    val = ((neurons.activity-neurons.TH-nox+sig >= 0.0) + 0.0)

    if hasattr(neurons, 'refractory_counter'):
        return val * (neurons.refractory_counter < 0.1)
    else:
        return val
    #neurons.output_new *= neurons.refractory_counter < 0.1

def last_cycle_step(neurons):
    return neurons.iteration % neurons.iteration_lag == neurons.iteration_lag-1

def first_cycle_step(neurons):
    return neurons.iteration % neurons.iteration_lag == 0




class SORN_init_afferent_synapses(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('init_afferent_synapses')

        #add synapsegroup.group_weighting attribute to change its weight compared to synapsegroups of same kind

        self.transmitter = self.get_init_attr('transmitter', None, neurons)

        density = self.get_init_attr('density', None, neurons)
        if type(density)==str and density[-1] == '%':
            density=float(density[:-1])/100

        density_mode = self.get_init_attr('density_mode', 'auto', neurons)

        self.distribution = self.get_init_attr('distribution', 'random_sample(dim)', neurons, do_not_diversify=True)

        if self.distribution is not None and ';plot' in self.distribution:
            import matplotlib.pyplot as plt
            self.distribution = self.distribution.replace(';plot', '')
            plt.hist(eval(self.distribution.replace(')', ',size=1000)')), bins=100)
            plt.show()

        partition_compensation = self.get_init_attr('partition_compensation', False, neurons)

        # np.random. ...
        #lognormal([mean, sigma, size])
        #normal([loc, scale, size])
        #random_sample(dim)
        #https://docs.scipy.org/doc/numpy-1.13.0/reference/routines.random.html


        if density == 'full':
            density_mode = 'full'
            density = 1

        if density_mode == 'auto':
            if np.min(density) < 1: #min because of array possibility
                density_mode = 'chance'
            else:
                density_mode = 'syn_count'


        if density_mode == 'full':
            density = 1
        elif density_mode == 'chance':
            density = density
        elif density_mode == 'syn_count':
            base_density = density#...
        elif density_mode == 'fixed_syn_count':
            density = -1
            #todo:implement (difficult with s.enabled...)
            #s.get_random_synapse_mat_fixed(syn_lamb)
        else:
            print('invalid density mode')

        self.group_weighting = self.get_init_attr('group_weighting', False, neurons)

        for s in neurons.afferent_synapses[self.transmitter]:
            if self.group_weighting == False:
                weighting = 1
            else:
                weighting = s.get_synapse_group_size_factor(s, self.transmitter)#todo test

            if partition_compensation:
                pc = s.src.group_without_subGroup().size/s.src.size
                weighting *= pc

            if density_mode == 'syn_count':
               density = base_density/s.src.size

            if self.distribution is not None:
                s.W = s.get_random_synapse_mat(density=float(density*weighting), rnd_code=self.distribution)#density*weighting
            else:
                s.W = (s.get_synapse_mat()+1)*s.enabled #all same

            s.enabled *= (s.W > 0)

            s.slow_add=0
            s.fast_add=0

        self.normalize = self.get_init_attr('normalize', True, neurons)# can be number or bool
        if self.normalize is not None and self.normalize is not False:
            if self.normalize==True:
                self.normalize=1.0
            normalize_synapse_attr('W', 'W', self.normalize, neurons, self.transmitter)

        #for s in neurons.afferent_synapses[self.transmitter]:
        #    plt.hist(s.W.flatten()[s.W.flatten()>0],bins=100)
        #    plt.show()


    def new_iteration(self, neurons):
        return

class SORN_init_neuron_vars(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('init_neuron_vars')

        af=self.get_init_attr('activation_function', 'binary', neurons)
        if af=='binary':
            neurons.activation_function = binary_activation_function
        elif af=='analog':
            neurons.activation_function = analog_activation_function
        else:
            neurons.activation_function=eval(af)
            #print('warning: no activation function defined')

        neurons.activity = neurons.get_neuron_vec()
        neurons.excitation = neurons.get_neuron_vec()
        neurons.inhibition = neurons.get_neuron_vec()
        neurons.input_act = neurons.get_neuron_vec()

        #neurons.slow_act=neurons.get_neuron_vec()

        neurons.TH = neurons.get_neuron_vec()+self.get_init_attr('init_TH', 0.5, neurons)#neurons.get_random_neuron_vec()*0.5

        neurons.iteration_lag = self.get_init_attr('iteration_lag', 1, neurons)
        neurons.rnd_act_factor = self.get_init_attr('rnd_act_factor', None, neurons)#sigma

    def new_iteration(self, neurons):
        if first_cycle_step(neurons):
            neurons.activity.fill(0)# *= 0
            neurons.excitation.fill(0)# *= 0
            neurons.inhibition.fill(0)# *= 0
            neurons.input_act.fill(0)# *= 0

            #neurons.slow_act *= 0

class SORN_NOX(Instant_Homeostasis):

    def partition_sum(self, neurons):
        neurons._temp_act_sum = neurons.get_neuron_vec()
        for s in neurons.afferent_synapses['GLU']:
            s.dst._temp_act_sum += np.mean(s.src.output_new)
        return neurons._temp_act_sum

    def set_variables(self, neurons):
        self.add_tag('NOX')
        super().set_variables(neurons)
        neurons.nox = neurons.get_neuron_vec()
        self.measurement_param = self.get_init_attr('mp', 'self.partition_sum(n)', None)#'np.mean(n.output_new)'
        self.adjustment_param = 'nox'
        self.measure_group_sum = True

        self.set_threshold(self.get_init_attr('h_dh', 0, neurons))#'same(IPTI, th)'
        self.adj_strength = -self.get_init_attr('eta_nox', 0.002, neurons)

        self.distance_sensitive=True

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.nox.fill(0)# *= 0
            super().new_iteration(neurons)

            #print(self.measurement_param)
            #neurons.nox *= 0.5#0.9


class SORN_signal_propagation_base(Behaviour):

    def set_variables(self, neurons):
        self.transmitter = self.get_init_attr('transmitter', None, neurons)
        self.strength = self.get_init_attr('strength', 1, neurons)  # 1 or -1

        if self.transmitter==None:
            print('warning no transmitter defined')
        #elif self.transmitter is not 'GLU' and self.transmitter is not 'GABA':
        #    print('warning unknown transmitter')

        if self.transmitter=='GLU' and self.strength<0:
            print('warning glutamate strength is inhibitory')

        if self.transmitter=='GABA' and self.strength>0:
            print('warning GABA strength is excitatory')

    def new_iteration(self, neurons):
        print('warning: signal_propagation_base has to be overwritten')


class SORN_external_input(NeuronActivator):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.strength = self.get_init_attr('strength', 1.0, neurons)
        neurons.input = np.zeros(neurons.size)
        self.write_to = 'input'
        neurons.add_tag('text_input_group')


    def new_iteration(self, neurons):
        super().new_iteration(neurons)
        if self.strength!=0:
            add = neurons.input * self.strength / neurons.iteration_lag
            neurons.activity += add

            #neurons.excitation += add #Todo:test

            neurons.input.fill(0)# *= 0
            neurons.input_act += add


class SORN_slow_syn(SORN_signal_propagation_base):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('slow_' + self.transmitter)
        self.sparse_optimization = self.get_init_attr('so', False, neurons)

    def new_iteration(self, neurons):
        if self.strength!=0:
            for s in neurons.afferent_synapses[self.transmitter]:
                if self.sparse_optimization:
                    s.slow_add = (np.sum(s.W[:, s.src.output.astype(np.bool)], axis=1) * self.strength) / neurons.iteration_lag
                    #add = (np.sum(s.W.transpose()[s.src.output.astype(np.bool)], axis=0) * self.strength) / neurons.iteration_lag
                else:
                    s.slow_add = (s.W.dot(s.src.output) * self.strength) / neurons.iteration_lag

                #neurons.slow_act += add

                s.dst.activity += s.slow_add
                if self.strength>0:
                    s.dst.excitation+=s.slow_add
                else:
                    s.dst.inhibition+=s.slow_add

        #if self.input_strength is not None:
        #    add = neurons.input * self.input_strength / neurons.iteration_lag
        #    neurons.activity += add

        #    #neurons.excitation += add #Todo:test
        #    neurons.input *= 0
        #    neurons.input_act += add



class SORN_fast_syn(SORN_signal_propagation_base):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('fast_' + self.transmitter)
        self.sparse_optimization = self.get_init_attr('so', False, neurons)

    def new_iteration(self, neurons):
        if self.strength!=0:
            for s in neurons.afferent_synapses[self.transmitter]:
                if self.sparse_optimization:
                    s.fast_add = (np.sum(s.W[:, s.src.activation_function(s.src).astype(np.bool)], axis=1) * self.strength) / neurons.iteration_lag
                    #add = (np.sum(s.W.transpose()[s.src.output.astype(np.bool)], axis=0) * self.strength) / neurons.iteration_lag
                else:
                    s.fast_add = s.W.dot(s.src.activation_function(s.src)) * self.strength / neurons.iteration_lag

                s.dst.activity += s.fast_add
                if self.strength > 0:
                    s.dst.excitation += s.fast_add
                else:
                    s.dst.inhibition += s.fast_add


class SORN_slow_input_collect(Behaviour):#has to be executed AFTER intra, inter...behaviours

    def set_variables(self, neurons):
        self.add_tag('input_collect')
        neurons.output_new_temp = neurons.get_neuron_vec() #required for STDP todo:remove

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.output_new_temp = neurons.activation_function(neurons)
            #reset moved to init_vars

            #neurons.test_val = (activation_function(neurons.activity, neurons) + activation_function(neurons.slow_act, neurons))/2


class SORN_input_collect(Behaviour):#has to be executed AFTER intra, inter...behaviours

    def set_variables(self, neurons):
        self.add_tag('input_collect')
        neurons.output = neurons.get_neuron_vec()
        neurons.output_new = neurons.get_neuron_vec() #required for STDP todo:remove

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.output_new = neurons.activation_function(neurons)
            #reset moved to init_vars

            #neurons.test_val = (activation_function(neurons.activity, neurons) + activation_function(neurons.slow_act, neurons))/2


class SORN_Refractory(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Refractory')
        neurons.refractory_counter = neurons.get_neuron_vec()
        self.factor = self.get_init_attr('factor', 0.9, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            #neurons.output_new *= neurons.refractory_counter < 0.1
            neurons.refractory_counter *= self.factor
            neurons.refractory_counter += neurons.output_new

class SORN_STDP_new(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('STDP')
        neurons.eta_stdp = self.get_init_attr('eta_stdp', 0.005, neurons)
        neurons.excitation_punishment = self.get_init_attr('excitation_punishment', 1, neurons)
        neurons.prune_stdp = self.get_init_attr('prune_stdp', False, neurons)


        for s in neurons.afferent_synapses['GLU']:
            s.STDP_src_lag_buffer_old = np.zeros(s.src.size)#neurons.size bug?
            s.STDP_src_lag_buffer_new = np.zeros(s.src.size)#neurons.size bug?

    def new_iteration(self, neurons):
        for s in neurons.afferent_synapses['GLU']:
            grow=s.dst.output_new[:, None] * s.src.output[None, :]
            shrink=neurons.excitation_punishment * s.dst.excitation[:, None] * s.src.output[None, :]
            exc_shrink=s.dst.output[:, None] * s.src.output_new[None, :]

            #print(np.sum(grow),np.sum(shrink),np.sum(exc_shrink))

            dw = neurons.eta_stdp * (grow - shrink - exc_shrink)

            s.W += dw * s.enabled
            s.W[s.W < 0.0] = 0.0


class SORN_STDP(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('STDP')
        neurons.eta_stdp = self.get_init_attr('eta_stdp', 0.005, neurons)
        neurons.prune_stdp = self.get_init_attr('prune_stdp', False, neurons)

        for s in neurons.afferent_synapses['GLU']:
            s.STDP_src_lag_buffer_old = np.zeros(s.src.size)#neurons.size bug?
            s.STDP_src_lag_buffer_new = np.zeros(s.src.size)#neurons.size bug?

    def new_iteration(self, neurons):
        for s in neurons.afferent_synapses['GLU']:

            #print(s.src.size, len(s.STDP_src_lag_buffer_new), len(s.src.output_new))

            s.STDP_src_lag_buffer_new += s.src.output_new/neurons.iteration_lag

            #buffer previous states

            if last_cycle_step(neurons):

                from_old = s.STDP_src_lag_buffer_old#
                from_new = s.STDP_src_lag_buffer_new#s.src.x_new#

                to_old = s.dst.output
                to_new = s.dst.output_new

                dw = neurons.eta_stdp * (to_new[:, None] * from_old[None, :] - to_old[:, None] * from_new[None, :])

                s.W += dw * s.enabled
                s.W[s.W < 0.0] = 0.0

                if neurons.prune_stdp:
                    s.W[s.W < 1e-10] = 0
                    s.enabled *= (s.W > 0)

                # print('STDP',s.W.sum())

                #clear buffer
                s.STDP_src_lag_buffer_old = s.STDP_src_lag_buffer_new.copy()
                s.STDP_src_lag_buffer_new = np.zeros(s.src.size)#neurons.size bug?



class SORN_SN(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('SN')
        self.syn_type = self.get_init_attr('syn_type', 'GLU', neurons)
        self.clip_max = self.get_init_attr('clip_max', None, neurons)
        neurons.weight_norm_factor = neurons.get_neuron_vec()+self.get_init_attr('init_norm_factor', 1.0, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            normalize_synapse_attr('W', 'W', neurons.weight_norm_factor, neurons, self.syn_type)
            for s in neurons.afferent_synapses[self.syn_type]:
                s.W=np.clip(s.W, 0, self.clip_max)


class SORN_iSTDP(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('iSTDP')

        self.transmitter = self.get_init_attr('transmitter', 'GABA', neurons)

        neurons.eta_istdp = self.get_init_attr('eta_istdp', 0.001, neurons)
        neurons.h_ip = self.get_init_attr('h_ip', 0.1, neurons)

        for s in neurons.afferent_synapses[self.transmitter]:
            s.iSTDP_src_lag_buffer = np.zeros(s.src.size)

    def new_iteration(self, neurons):
        for s in neurons.afferent_synapses[self.transmitter]:
            s.iSTDP_src_lag_buffer += s.src.output/neurons.iteration_lag
            if last_cycle_step(neurons):

                s.W += -s.dst.eta_istdp * ((1 - (s.dst.excitation+s.dst.inhibition) * (1 + 1.0 / s.dst.h_ip))[:, None] * s.iSTDP_src_lag_buffer[None, :])
                s.W = np.clip(s.W, 0.001, 1.0)*s.enabled

                s.iSTDP_src_lag_buffer.fill(0)# *= 0

                #dW_same = -self.eta_istdp * ((1 - (s.dst.output[:, None] * (1 + 1.0 / self.h_ip))) * s.src.output[None, :])


class SORN_SC_TI(Time_Integration_Homeostasis):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('SCTI')
        self.measurement_param = self.get_init_attr('mp', 'n.excitation', neurons)
        self.adjustment_param = 'weight_norm_factor'

        self.set_threshold(self.get_init_attr('h_sc', 0.1, neurons))
        self.adj_strength = self.get_init_attr('eta_sc', 0.001, neurons)

        self.target_clip_min = 0.5


    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            super().new_iteration(neurons)

class SORN_IP_TI(Time_Integration_Homeostasis):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('IPTI')
        self.measurement_param = self.get_init_attr('mp', 'n.output_new', neurons)
        self.adjustment_param = 'TH'

        self.set_threshold(self.get_init_attr('h_ip', 0.1, neurons))
        self.adj_strength = -self.get_init_attr('eta_ip', 0.001, neurons)
        self.target_clip_min=self.get_init_attr('clip_min', -0.001, neurons)
        #target_clip_max

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            super().new_iteration(neurons)



class SORN_diffuse_IP(Time_Integration_Homeostasis):

    def set_variables(self, neurons):
        self.add_tag('diff_IP')
        super().set_variables(neurons)
        self.measurement_param = self.get_init_attr('mp', 'test_val', neurons)#output_new
        self.adjustment_param = 'TH'
        self.measure_group_sum = True

        h_dh=self.get_init_attr('h_dh', 0.1, neurons)

        #if h_dh == 'IPTI_h_ip':
        #    h_dh = np.mean((neurons['IPTI',0].min_th+neurons['IPTI',0].max_th)/2)

        self.set_threshold(h_dh)
        self.adj_strength = -self.get_init_attr('eta_dh', 0.001, neurons)


    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            super().new_iteration(neurons)



class SORN_finish(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('finish')

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.output = neurons.output_new
            #neurons.exc=neurons.excitation.copy()


class SORN_random_act(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('rnd_act')
        self.chance = self.get_init_attr('chance', 10.0, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            e = np.mean(neurons.excitation)
            c = self.chance*np.abs(0.005-e)
            add=0.1*(neurons.get_random_neuron_vec()<c)
            neurons.excitation += add
            neurons.activity += add






'''
class SORN_init_variables(Neuron_Behaviour):

    def set_variables(self, neurons):
        self.add_tag('init vars')
        neurons.iteration_lag = self.get_init_attr('iteration_lag', 1, neurons)

        neurons.require_synapses('GLU')
        neurons.require_synapses('GABA')

        neurons.inh_strength = self.get_init_attr('inh_strength', 1.0, neurons)

        neurons.x = (np.random.random(neurons.size) < 0.5) + 0
        neurons.T_min = self.get_init_attr('T_min', 0.0, neurons)
        neurons.T_max = self.get_init_attr('T_max', 0.5, neurons)
        neurons.T = neurons.T_min + np.random.random(neurons.size) * (neurons.T_max - neurons.T_min)
        neurons.input_gain = self.get_init_attr('input_gain', 1.0, neurons)
        neurons.sigma = self.get_init_attr('sigma', None, neurons)
        neurons.R_x = np.zeros(neurons.size)
        neurons.lamb = self.get_init_attr('lamb', -1, neurons)#10 -1=full
        neurons.x_new = np.zeros(neurons.size)
        neurons.input = np.zeros(neurons.size)

        neurons.FSC = self.get_init_attr('fixed_synapse_count', False, neurons)

        #todo: calculate distributed lamb for every synapse group

        for s in neurons.afferent_synapses['GLU']:
            syn_lamb = neurons.lamb#*s.get_synapse_group_size_factor(neurons, 'GLU')
            if neurons.lamb == -1:
                syn_lamb = neurons.size

            if neurons.FSC:
                s.W = s.get_random_synapse_mat_fixed(syn_lamb)*s.enabled
            else:
                s.W = s.get_random_synapse_mat(density=syn_lamb / float(neurons.size))*s.enabled
            s.enabled *= (s.W > 0)


        for s in neurons.afferent_synapses['GLU']:
            if 'sparse' in s.tags:
                s.W = sp.csc_matrix(s.W)
                #s.W.eliminate_zeros()

        for s in neurons.afferent_synapses['GABA']:
            s.W = s.get_random_synapse_mat()

        normalize_synapse_attr('W', 'W', 1.0, neurons, 'GLU')
        normalize_synapse_attr('W', 'W', 1.0, neurons, 'GABA')

    def new_iteration(self, neurons):
        return
'''



from PymoNNto.Exploration.Network_UI.TabBase import *

class scatter_tab(TabBase):

    def __init__(self, x_var, y_var, title='Scatter', timesteps=500):
        super().__init__(title)
        self.x_var = x_var
        self.y_var = y_var
        self.timesteps=timesteps

    def add_recorder_variables(self, neuron_group, Network_UI):
        if hasattr(neuron_group, self.x_var):
            Network_UI.add_recording_variable(neuron_group, 'n.' + self.x_var, timesteps=self.timesteps)
        if hasattr(neuron_group, self.y_var):
            Network_UI.add_recording_variable(neuron_group, 'n.' + self.y_var, timesteps=self.timesteps)

    def initialize(self, Network_UI):
        self.scatter_tab = Network_UI.Next_Tab(self.title)

        p = Network_UI.Add_plot(x_label=self.x_var, y_label=self.y_var)
        self.scatter_items=[]
        for i, _ in enumerate(Network_UI.neuron_visible_groups):
            spi = pg.ScatterPlotItem()
            self.scatter_items.append(spi)

            p.addItem(spi)
            #Network_UI.Next_H_Block()

    def update(self, Network_UI):
        if self.scatter_tab.isVisible():
            for i, group_tag in enumerate(Network_UI.neuron_visible_groups):
                if len(Network_UI.network[group_tag]) > 0:
                    group = Network_UI.network[group_tag, 0]
                    #group=Network_UI.network[Network_UI.neuron_select_group, 0]

                    if hasattr(group, self.x_var) and hasattr(group, self.y_var):
                        #rec = Network_UI.rec(group, self.timesteps)
                        x_values = group['n.'+self.x_var, 0, 'np'][-self.timesteps:]
                        y_values = group['n.'+self.y_var, 0, 'np'][-self.timesteps:]

                        x_val = np.mean(x_values, axis=0)
                        y_val = np.mean(y_values, axis=0)

                        self.scatter_items[i].setData(x_val.copy(), y_val.copy(), brush=group.color)



from PymoNNto.Exploration.Network_UI.Basic_Tabs.sidebar_activity_module import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.sidebar_fast_forward_module import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.sidebar_save_load_module import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.multi_group_plot_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.hist_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.info_tabs import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.single_group_plot_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.stability_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.weight_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.fourier_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.partition_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.spiketrain_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.individual_weight_tab import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.PCA_tab import *


def get_default_UI_modules(neuron_parameters=['output'], synapse_parameters=['W']):
    return [
    UI_sidebar_activity_module(1, add_color_dict={neuron_parameters[0]:(255, 255, 255)}),
    multi_group_plot_tab(neuron_parameters),#['output', 'TH', 'weight_norm_factor', 'nox', 'refractory_counter']
    spiketrain_tab(parameter=neuron_parameters[0]),
    weight_tab(weight_attrs=synapse_parameters),#, 'W_temp', 'W_stable'
    partition_tab(),
    PCA_tab(parameter=neuron_parameters[0]),
    #individual_weight_tab(),
    hist_tab(weight_attr=synapse_parameters[0]),
    single_group_plot_tab({neuron_parameters[0]:(0, 0, 0)}),#, 'excitation':(0, 0, 255), 'inhibition':(255, 0, 0), 'input_act':(255, 0, 255), 'TH':(0, 255, 0)}),
    stability_tab(parameter=neuron_parameters[0]),
    fourier_tab(parameter=neuron_parameters[0]),
    info_tab(),
    sidebar_fast_forward_module(),
    sidebar_save_load_module()
]


from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.NetworkCore.Base import *
from PymoNNto.Exploration.StorageManager.StorageManager import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.info_tabs import *
import copy
import pickle
import os

compile_class = type(compile('1+1', '<string>', 'eval'))
base_obj_type = NetworkObjectBase

class sidebar_save_load_module(TabBase):

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def clear_compiled_code(self, obj, visited=[]):

        if isinstance(obj, base_obj_type) and (obj not in visited):
            visited.append(obj)
            d = obj.__dict__
            for key in d:
                if type(getattr(obj, key)) in [list, dict] or isinstance(getattr(obj, key), base_obj_type):
                    self.clear_compiled_code(d[key], visited)

                if type(getattr(obj, key)) is compile_class:
                    setattr(obj, key, None)

        visited.append(obj)

        if type(obj) is list:
            for i in range(len(obj)):
                if type(obj[i]) in [list, dict] or isinstance(obj[i], base_obj_type):
                    self.clear_compiled_code(obj[i], visited)

                if type(obj[i]) is compile_class:
                    obj[i] = None

        if type(obj) is dict:
            for sub_obj_key in obj:
                if type(obj[sub_obj_key]) in [list, dict] or isinstance(obj[sub_obj_key], base_obj_type):
                    self.clear_compiled_code(obj[sub_obj_key], visited)

                if type(obj[sub_obj_key]) is compile_class:
                    obj[sub_obj_key] = None




    def initialize(self, Network_UI):

        self.folder=get_data_folder()+'/NetworkStates/'

        if not os.path.exists(self.folder):
            try:
                os.mkdir(self.folder)
            except:
                print('was not able to create Data/NetworkStates folder')

        h_layout = Network_UI.Add_Sidebar_Element(return_h_layout=True)

        self.save_edit = QLineEdit('name...')
        h_layout.addWidget(self.save_edit)

        def save(event):
            net = copy.deepcopy(Network_UI.network)
            self.clear_compiled_code(net)
            pickle.dump(net, open(self.folder + self.save_edit.text() + '.netstate', 'wb'))
            self.load_box.addItem(self.save_edit.text())

        self.save_btn = QPushButton('save', Network_UI.main_window)
        self.save_btn.clicked.connect(save)
        h_layout.addWidget(self.save_btn)

        #h_layout = Network_UI.Add_Sidebar_Element(return_h_layout=True)

        self.load_box = QComboBox()
        for file in os.listdir(self.folder):
            if '.obj' in str(file):
                self.load_box.addItem(file.replace('.netstate', ''))
        h_layout.addWidget(self.load_box)

        def load(event):
            Network_UI.network = pickle.load(open(self.folder + self.load_box.currentText() + '.netstate', 'rb'))

            for tab in Network_UI.infotabs:
                Network_UI.tabs.removeTab(len(Network_UI.tabs) - 1)

            Network_UI.modules.append(info_tab(Network_UI))

        self.load_btn = QPushButton('load', Network_UI.main_window)
        self.load_btn.clicked.connect(load)
        h_layout.addWidget(self.load_btn)



    def update(self, Network_UI):
        return


#from mnist import MNIST #pip install python-mnist
import numpy as np
from PIL import Image, ImageDraw, ImageFilter, ImageEnhance


def rotatearoundpoint(p, rot, middle):
    aa = p[0] - middle[0]
    bb = p[1] - middle[1]
    cc = np.sqrt(aa * aa + bb * bb)
    a = 0
    if cc != 0:
        a = np.rad2deg(np.arcsin(aa / cc))
    if bb < 0:
        a = 180 - a
    a = np.deg2rad(a + rot)
    aa = np.sin(a) * cc
    bb = np.cos(a) * cc
    return (middle[0] + aa, middle[1] + bb)


def picture_to_array(image, max=1):
    pil_image_gray = image.convert('L')
    result = np.array(pil_image_gray).astype(np.float64)
    return result/np.maximum(max, np.max(image))


#def picture_to_array(image, max=1):
#    result = []
#    for y in range(image.size[1]):
#        for x in range(image.size[0]):
#            val = image.getpixel((x, y))
#            result.append(val)
#            if val > max:
#                max = val
#    for i, val in enumerate(result):
#        result[i] = val / max
#    return result


def getRotLinePicture(deg, width=15, height=15):
    im = Image.new('L', (width, height), (0))
    draw = ImageDraw.Draw(im)
    rot_point = rotatearoundpoint((width / 2, 0), deg, (0, 0))
    draw.line(
        (width / 2 - rot_point[0], height / 2 - rot_point[1], width / 2 + rot_point[0], height / 2 + rot_point[1]),
        fill=255)
    return picture_to_array(im)


def getVerLinePicture(x, width=15, height=15):
    im = Image.new('L', (width, height), (0))
    draw = ImageDraw.Draw(im)
    draw.line((x, 0, x, height - 0), fill=255)
    return picture_to_array(im)
    # im.show()
    # im.save("Data/im{}.png".format(deg))


def getVerLinePictureSmooth(x, width=15, height=15):
    im = Image.new('L', (width, height), (0))
    draw = ImageDraw.Draw(im)
    for i in [-2, -1, 0, 1, 2]:
        shade = 255 - abs(i) * 80
        draw.line((x + i, 0, x + i, height - 0), fill=shade)
    return picture_to_array(im)
    # im.show()
    # im.save("Data/im{}.png".format(deg))


def getHorLinePicture(y, width=15, height=15):
    im = Image.new('L', (width, height), (0))
    draw = ImageDraw.Draw(im)
    draw.line((0, y, width - 0, y), fill=255)
    return picture_to_array(im)
    # im.show()
    # im.save("Data/im{}.png".format(deg))


def getHorLinePictureSmooth(y, width=15, height=15):
    im = Image.new('L', (width, height), (0))
    draw = ImageDraw.Draw(im)
    for i in [-2, -1, 0, 1, 2]:
        shade = 255 - abs(i) * 80
        draw.line((0, y + i, width - 0, y + i), fill=shade)
    return picture_to_array(im)


def getLinePicture(deg, center_x, center_y, length, width, height):
    im = Image.new('L', (width, height), (0))
    draw = ImageDraw.Draw(im)
    rot_point = rotatearoundpoint((length / 2, 0), deg, (0, 0))
    draw.line((center_x - np.floor(rot_point[0]), center_y - np.floor(rot_point[1]), center_x + np.floor(rot_point[0]), center_y + np.floor(rot_point[1])), fill=255)
    return picture_to_array(im)


def get_rgbw(r, g, b, w):
    result = np.zeros(r.shape)
    for x in range(int(r.shape[1] / 2)):
        for y in range(int(r.shape[0] / 2)):
            x2 = x * 2
            y2 = y * 2
            result[y2, x2] = np.sum(r[y2:y2 + 2, x2:x2 + 2]) / 4
            result[y2 + 1, x2] = np.sum(g[y2:y2 + 2, x2:x2 + 2]) / 4
            result[y2, x2 + 1] = np.sum(b[y2:y2 + 2, x2:x2 + 2]) / 4
            result[y2 + 1, x2 + 1] = np.sum(w[y2:y2 + 2, x2:x2 + 2]) / 4

    return result


def get_log_kenel(sigma):
    size = int(2 * (np.ceil(3 * sigma)) + 1)
    x, y = np.meshgrid(np.arange(-size / 2, size / 2 + 1), np.arange(-size / 2, size / 2 + 1))
    normal = 1 / (2.0 * np.pi * sigma ** 2)
    kernel = ((x ** 2 + y ** 2 - (2.0 * sigma ** 2)) / sigma ** 4) * np.exp(
        -(x ** 2 + y ** 2) / (2.0 * sigma ** 2)) / normal
    # fig = plt.figure()
    # ax = fig.add_subplot(111, projection='3d')
    # zs = kernel#np.array(fun(np.ravel(X), np.ravel(Y)))
    # ax.plot_surface(x, y, zs)
    # ax.set_xlabel('X Label')
    # ax.set_ylabel('Y Label')
    # ax.set_zlabel('Z Label')
    # plt.show()
    return kernel


def get_multi_color_LOG_On_Off(c1, c2):
    from scipy import signal
    kernel = get_log_kenel(1)
    p_kernel = kernel * (kernel > 0)
    n_kernel = kernel * (kernel < 0)

    c1_p = signal.convolve2d(c1, p_kernel, boundary='symm', mode='same')  # center
    c2_n = signal.convolve2d(c2, n_kernel, boundary='symm', mode='same')  # surround

    LOG = c1_p + c2_n
    on_center = (LOG > 0) * LOG
    off_center = (LOG < 0) * LOG * -1

    return on_center.astype(np.float64), off_center.astype(np.float64)


def get_LOG_On_Off(data, sigma=1):
    import scipy.ndimage as ndimage
    LOG = ndimage.gaussian_laplace(data.astype(np.int32), sigma=sigma)
    on_center = (LOG > 0) * LOG
    off_center = (LOG < 0) * LOG * -1
    return on_center, off_center


def gkern(kernlen=9, nsig=2):
    import scipy.ndimage.filters as fi
    #print(kernlen, nsig)
    inp = np.zeros((kernlen, kernlen))
    inp[kernlen // 2, kernlen // 2] = 1
    return fi.gaussian_filter(inp, nsig)


from time import time
import numpy as np
import random
#from NetworkBehaviour.Input.Activator import *
from PymoNNto.NetworkCore.Base import *
from PymoNNto.NetworkCore.Synapse_Group import *
from PymoNNto.Exploration.Evolution.Interface_Functions import *
import copy

class Network(NetworkObjectBase):

    def __init__(self, tag=None, behaviour={}):
        super().__init__(tag)

        self.NeuronGroups = []
        self.SynapseGroups = []

        self.behaviour = behaviour

        self.iteration = 0


    def set_mechanisms(self, keys, enabeled):
        for key in keys:
            for obj in self.all_behaviour_objects():
                for mechansim in obj[key]:
                    mechansim.active = enabeled


    def deactivate_mechanisms(self, keys):
        self.set_mechanisms(keys, False)

    def activate_mechanisms(self, keys):
        self.set_mechanisms(keys, True)

    def recording_off(self):
        for obj in self.all_behaviour_objects():
            obj.recording = False

    def recording_on(self):
        for obj in self.all_behaviour_objects():
            obj.recording = True

    def all_behaviour_objects(self):
        return [self]+self.NeuronGroups+self.SynapseGroups

    def clear_recorder(self, keys=None):
        for obj in self.all_behaviour_objects():
            for key in obj.behaviour:
                if (keys is None or key in keys) and hasattr(obj.behaviour[key], 'clear_recorder'):
                    obj.behaviour[key].clear_recorder()

    def set_marked_variables(self, info=True, storage_manager=None):
        current_genome={}
        for obj in self.all_behaviour_objects():
            for key in obj.behaviour:
                b=obj.behaviour[key]
                for variable_key in b.init_kwargs:
                    while type(b.init_kwargs[variable_key]) is str and '[' in b.init_kwargs[variable_key] and ']' in b.init_kwargs[variable_key]:
                        s = b.init_kwargs[variable_key]

                        start = s.index('[')
                        end = s.index(']')
                        internal = s[start+1: end].split('#')
                        default_value = float(internal[0])
                        gene_key = internal[1]

                        current_genome[gene_key] = get_gene(gene_key, default_value)

                        b.init_kwargs[variable_key] = s[:start]+'{:.15f}'.format(current_genome[gene_key]).rstrip('0').rstrip('.')+s[end+1:]
        if info:
            print('default genome:', current_genome)

        if storage_manager is not None:
            storage_manager.save_param(key='evolution_params', value=current_genome)

    def find_objects(self, key):
        result = []

        if key in self.behaviour:
            result.append(self.behaviour)

        for bk in self.behaviour:
            behaviour = self.behaviour[bk]
            result += behaviour[key]

        for ng in self.NeuronGroups:
            result += ng[key]

        for sg in self.SynapseGroups:
            result += sg[key]

        return result

    def print_net_info(self):
        neuron_count = np.sum(np.array([ng.size for ng in self.NeuronGroups]))
        sysnape_count = np.sum(np.array([sg.src.size*sg.dst.size for sg in self.SynapseGroups]))

        print('initialize tren... Neurons: ', neuron_count, '|', len(self.NeuronGroups), ' blocks, Synapses: ', sysnape_count, '|', len(self.SynapseGroups),' blocks')

    def initialize(self, evo_replace_param_list=None, info=False, warnings=True, storage_manager=None):

        self.set_marked_variables(info=info, storage_manager=storage_manager)

        if info:
            self.print_net_info()

        if evo_replace_param_list is not None:
            self.set_marked_variables(evo_replace_param_list)

        #self.old_param_list = self.get_all_params()

        self.set_synapses_to_neuron_groups()

        self.behaviour_timesteps = []

        self.add_behaviour_keys_dict(self.behaviour)# todo: test

        for ng in self.NeuronGroups:
            self.add_behaviour_keys_dict(ng.behaviour)

        for sg in self.SynapseGroups:
            self.add_behaviour_keys_dict(sg.behaviour)

        self.set_variables()
        #self.save_network_state()

        self.check_unique_tags(warnings)

    def check_unique_tags(self,warnings=True):
        unique_tags=[]
        for ng in self.NeuronGroups:

            if len(ng.tags) == 0:
                ng.tags.append('NG')
                print('no tag defined for NeuronGroup. "NG" tag added')

            if ng.tags[0] in unique_tags:
                counts=unique_tags.count(ng.tags[0])
                new_tag=ng.tags[0]+chr(97+counts)
                unique_tags.append(ng.tags[0])
                if warnings:
                    print('Warning: NeuronGroup Tag "'+ng.tags[0]+'" already in use. The first Tag of an Object should be unique and will be renamed to "'+new_tag+'". Multiple Tags can be sperated with a "," (NeuronGroup(..., tag="tag1,tag2,..."))')
                ng.tags[0] = new_tag

            else:
                unique_tags.append(ng.tags[0])

        #for sg in self.SynapseGroups:
        #    if sg.tags[0] in unique_tags:
        #        print('Warning: NeuronGroup Tag "' + sg.tags[0] + '" already in use. The first Tag of an Object should be unique. Multiple Tags can be sperated with a "," (SynapseGroup(..., tag="tag1,tag2,..."))')
        #    unique_tags.append(sg.tags[0])

    def add_behaviour_key(self, key):
        self.behaviour_timesteps.append(key)
        self.behaviour_timesteps.sort()

    def add_behaviour_keys_dict(self, behaviour_dict):
        for key in behaviour_dict:
            if not key in self.behaviour_timesteps:
                self.add_behaviour_key(key)

    def add_behaviours_to_objects(self, behaviours, net_objs):
        for obj in net_objs:
            self.add_behaviours_to_object(copy.copy(behaviours), obj)

    def add_behaviours_to_object(self, behaviours, net_obj):
        self.clear_tag_cache()
        original = behaviours
        #if type(behaviours) == list:
        #    start_key = self.behaviour_timesteps[-1]
        #    d = {}
        #    for b in behaviours:
        #        start_key += 1
        #        d[start_key] = b
        #    behaviours = d

        for key in behaviours:
            net_obj.behaviour[key] = behaviours[key]
            net_obj.behaviour[key].set_variables(net_obj)
            net_obj.behaviour[key].check_unused_attrs()


        self.add_behaviour_keys_dict(behaviours)
        return original

    def remove_behaviours_from_objects(self, net_objs, keys=[], tags=[]):
        for obj in net_objs:
            self.remove_behaviours_from_object(obj, keys, tags)

    def remove_behaviours_from_object(self, net_obj, keys=[], tags=[]):
        found=[]
        found+=keys
        for key in net_obj.behaviour:
            for tag in tags:
                if tag in net_obj.behaviour[key].tags:
                    found.append(key)

        found=list(set(found))#unique

        for key in found:
            net_obj.behaviour.pop(key)

        self.clear_tag_cache()

    def clear_tag_cache(self):

        for obj in self.all_behaviour_objects():
            obj.clear_cache()

            for k in obj.behaviour:
                obj.behaviour[k].clear_cache()


    def set_variables(self):
        for timestep in self.behaviour_timesteps:

            for obj in self.all_behaviour_objects():

                if timestep in obj.behaviour:
                    if not obj.behaviour[timestep].run_on_init:
                        obj.behaviour[timestep].set_variables(obj)
                        obj.behaviour[timestep].check_unused_attrs()

            '''
            for ng in self.NeuronGroups:
                if timestep in ng.behaviour:
                    if not ng.behaviour[timestep].run_on_neuron_init_var:
                        ng.behaviour[timestep].set_variables(ng)
                        ng.behaviour[timestep].check_unused_attrs()

            for sg in self.SynapseGroups:
                if timestep in sg.behaviour:
                    if not sg.behaviour[timestep].run_on_synapse_init_var:
                        sg.behaviour[timestep].set_variables(sg)
                        sg.behaviour[timestep].check_unused_attrs()
            '''


    def set_synapses_to_neuron_groups(self):
        for ng in self.NeuronGroups:

            ng.afferent_synapses = {'All':[]}
            ng.efferent_synapses = {'All':[]}

            for sg in self.SynapseGroups:
                for tag in sg.tags:
                    ng.afferent_synapses[tag] = []
                    ng.efferent_synapses[tag] = []

            for sg in self.SynapseGroups:
                if sg.dst.BaseNeuronGroup == ng:
                    for tag in sg.tags+['All']:
                        ng.afferent_synapses[tag].append(sg)

                if sg.src.BaseNeuronGroup == ng:
                    for tag in sg.tags+['All']:
                        ng.efferent_synapses[tag].append(sg)


    def simulate_iteration(self, measure_behaviour_execution_time=False):

        if measure_behaviour_execution_time:
            time_measures={}

        self.iteration += 1
        for timestep in self.behaviour_timesteps:

            for net_obj in self.all_behaviour_objects():
                net_obj.iteration=self.iteration
                if timestep in net_obj.behaviour and net_obj.behaviour[timestep].behaviour_enabled:
                    if measure_behaviour_execution_time:
                        start_time = time()
                        net_obj.behaviour[timestep].new_iteration(net_obj)
                        time_measures[timestep] = (time() - start_time) * 1000
                    else:
                        net_obj.behaviour[timestep].new_iteration(net_obj)

        if measure_behaviour_execution_time:
            return time_measures

            '''
            if timestep in self.behaviour and self.behaviour[timestep].behaviour_enabled:
                if measure_behaviour_execution_time:
                    start_time = time()
                    self.network_behaviour[timestep].new_iteration(self)
                    time_measures[timestep] = (time() - start_time) * 1000
                else:
                    self.network_behaviour[timestep].new_iteration(self)

            for ng in self.NeuronGroups:
                ng.iteration=self.iteration
                if timestep in ng.behaviour and ng.behaviour[timestep].behaviour_enabled:
                    if measure_behaviour_execution_time:
                        start_time = time()
                        ng.behaviour[timestep].new_iteration(ng)
                        time_measures[timestep] = (time() - start_time) * 1000
                    else:
                        ng.behaviour[timestep].new_iteration(ng)

            for sg in self.SynapseGroups:
                sg.iteration=self.iteration
                if timestep in sg.behaviour and sg.behaviour[timestep].behaviour_enabled:
                    if measure_behaviour_execution_time:
                        start_time = time()
                        sg.behaviour[timestep].new_iteration(sg)
                        time_measures[timestep] = (time() - start_time) * 1000
                    else:
                        sg.behaviour[timestep].new_iteration(ng)
            '''


    def simulate_iterations(self, iterations, batch_size=-1, measure_block_time=False, disable_recording=False):

        if type(iterations) is str:
            iterations=self['Clock', 0].time_to_iterations(iterations)

        time_diff=None

        if disable_recording:
            self.recording_off()

        #if measure_block_time:
        #    print('batch size: {}'.format(block_iterations))

        if batch_size==-1:
            outside_it = 1
            block_iterations = iterations
        else:
            outside_it=int(iterations/batch_size)
            block_iterations=batch_size

        for t in range(int(outside_it)):
            if measure_block_time:
                start_time = time()
            for i in range(int(block_iterations)):
                self.simulate_iteration()
            if measure_block_time:
                time_diff = (time() - start_time) * 1000

                print('\r{}xBatch: {}/{} ({}%) {:.3f}ms'.format(block_iterations,t+1, outside_it, int(100/outside_it*(t+1)),time_diff), end='')#, end='')

        for i in range(iterations%batch_size):
            self.simulate_iteration()

        if disable_recording:
            self.recording_on()

        return time_diff


    #def partition_Synapse_Groups(self, SynapseGroups=[]):
    #    #todo: make sure that all mechanisms use "s.src" and "s.dst" and not "neurons"!
    #    if SynapseGroups==[]:
    #        SynapseGroups=self.SynapseGroups.copy()

    #    for sg in SynapseGroups:
    #        self.partition_Synapse_Group(sg)

    def partition_Synapse_Group3(self, synapse_group, steps):
        return self.partition_Synapse_Group2(synapse_group, synapse_group.dst.partition_steps(steps))

    def partition_Synapse_Group2(self, synapse_group, dst_groups):#todo:auto receptive field extraction (blocks dont need to be squared!)

        rf_x, rf_y, rf_z = synapse_group.get_max_receptive_field_size()

        syn_sub_groups=[]

        for dst_subgroup in dst_groups:

            src_x_start = np.min(dst_subgroup.x)-rf_x
            src_x_end = np.max(dst_subgroup.x)+rf_x

            src_y_start = np.min(dst_subgroup.y)-rf_y
            src_y_end = np.max(dst_subgroup.y)+rf_y

            src_z_start = np.min(dst_subgroup.z)-rf_z
            src_z_end = np.max(dst_subgroup.z)+rf_z

            src_mask = (synapse_group.src.x >= src_x_start) * (synapse_group.src.x <= src_x_end) * (synapse_group.src.y >= src_y_start) * (synapse_group.src.y <= src_y_end) * (synapse_group.src.z >= src_z_start) * (synapse_group.src.z <= src_z_end)

            sg=SynapseGroup(synapse_group.src.subGroup(src_mask), dst_subgroup, net=None)

            syn_sub_groups.append(sg)

            # partition enabled update
            if type(synapse_group.enabled) is np.ndarray:
                mat_mask = dst_subgroup.mask[:, None] * src_mask[None, :]
                sg.enabled = synapse_group.enabled[mat_mask].copy().reshape(sg.get_synapse_mat_dim())

            # copy al attributes
            sgd = synapse_group.__dict__
            for key in sgd:
                if key not in ['src', 'dst', 'enabled']:
                    setattr(sg, key, copy.copy(sgd[key]))

        #add sub Groups
        for sg in syn_sub_groups:
            self.SynapseGroups.append(sg)

        #remove original SG
        self.SynapseGroups.remove(synapse_group)

        return syn_sub_groups


    def partition_Synapse_Group(self, syn_group, receptive_field_size=1, split_size=1):

        src = syn_group.src
        dst = syn_group.dst

        #adjust param
        if type(split_size) is int:
            split_size = [split_size, split_size, split_size]

        if len(split_size) == 2:
            split_size.append(1)

        if type(receptive_field_size) is int:
            receptive_field_size = [receptive_field_size, receptive_field_size, receptive_field_size]

        if len(receptive_field_size) == 2:
            receptive_field_size.append(0)

        # check dimensions
        if hasattr(src, 'width'):
            src_min = [np.min(p) for p in [src.x, src.y, src.z]]
            src_max = [np.max(p) for p in [src.x, src.y, src.z]]
        else:
            src_min = [0, 0, 0]
            src_max = [src.size, 0, 0]

        if hasattr(dst, 'width'):
            dst_min = [np.min(p) for p in [dst.x, dst.y, dst.z]]
            dst_max = [np.max(p) for p in [dst.x, dst.y, dst.z]]
        else:
            dst_min = [0, 0, 0]
            dst_max = [dst.size, 0, 0]

        def get_start_end(step, dim, group='src'):
            if group == 'src':
                start=src_min[dim]+(src_max[dim]-src_min[dim])/split_size[dim]*step-receptive_field_size[dim]
                end=src_min[dim]+(src_max[dim]-src_min[dim])/split_size[dim]*(step+1)+receptive_field_size[dim]
            if group == 'dst':
                start=dst_min[dim]+(dst_max[dim]-dst_min[dim])/split_size[dim]*step
                end=dst_min[dim]+(dst_max[dim]-dst_min[dim])/split_size[dim]*(step+1)
            return start, end

        #calculate sub Groups
        sub_groups = []

        #src_masks = []
        dst_masks = []

        for w_step in range(split_size[0]):          #x_steps
            src_x_start, src_x_end = get_start_end(w_step, 0, 'src')
            dst_x_start, dst_x_end = get_start_end(w_step, 0, 'dst')
            for h_step in range(split_size[1]):      #y_steps
                src_y_start, src_y_end = get_start_end(h_step, 1, 'src')
                dst_y_start, dst_y_end = get_start_end(h_step, 1, 'dst')
                for d_step in range(split_size[2]):  #z_steps
                    src_z_start, src_z_end = get_start_end(d_step, 2, 'src')
                    dst_z_start, dst_z_end = get_start_end(d_step, 2, 'dst')

                    src_mask = (src.x >= src_x_start) * (src.x <= src_x_end) * (src.y >= src_y_start) * (src.y <= src_y_end) * (src.z >= src_z_start) * (src.z <= src_z_end)
                    dst_mask = (dst.x >= dst_x_start) * (dst.x <= dst_x_end) * (dst.y >= dst_y_start) * (dst.y <= dst_y_end) * (dst.z >= dst_z_start) * (dst.z <= dst_z_end)

                    #remove duplicates
                    #for old_src_mask in src_masks:
                    #    src_mask[old_src_mask] *= False

                    for old_dst_mask in dst_masks:
                        dst_mask[old_dst_mask] *= False

                    #print(np.sum(src_mask), np.sum(dst_mask))

                    #src_masks.append(src_mask)
                    dst_masks.append(dst_mask)

                    #import matplotlib.pyplot as plt
                    #plt.scatter(src.x, src.y, c=src_mask)
                    #plt.scatter(dst.x+50, dst.y, c=dst_mask)
                    #plt.show()

                    beh = {}
                    for k in syn_group.behaviour:
                        beh[k] = copy.copy(syn_group.behaviour[k])

                    sg = SynapseGroup(syn_group.src.subGroup(src_mask), syn_group.dst.subGroup(dst_mask), behaviour=beh)

                    #partition enabled update
                    if type(syn_group.enabled) is np.ndarray:
                        mat_mask = dst_mask[:, None] * src_mask[None, :]
                        sg.enabled = syn_group.enabled[mat_mask].copy().reshape(sg.get_synapse_mat_dim(), net=None)

                    #copy al attributes
                    sgd = syn_group.__dict__
                    for key in sgd:
                        if key not in ['behaviour', 'src', 'dst', 'enabled']:
                            setattr(sg, key, copy.copy(sgd[key]))

                    sub_groups.append(sg)

        #add sub Groups
        for sg in sub_groups:
            self.SynapseGroups.append(sg)

        #remove original SG
        self.SynapseGroups.remove(syn_group)

        return sub_groups









    #def learning_off(self):


    #def learning_on(self):
    #    for ng in self.NeuronGroups:
    #        ng.learning = True

    #def clear_recorder(self):
    #    for ng in self.NeuronGroups:
    #        for key in ng.behaviour:
    #            if isinstance(ng.behaviour[key], )
    #                rr.clear_recorder()


    #def save_network_state(self):
    #    self.new_param_list=self.get_all_params()
    #    self.group_value_dicts={}
    #    for group in self.old_param_list:#same group for both
    #        value_dict={}
    #        for key in self.new_param_list[group]:
    #            if not key in self.old_param_list:
    #                val=getattr(group, key)
    #                t = type(val)
    #                if t == np.ndarray or t == list:
    #                    value_dict[key]=val.copy()
    #                else:
    #                    value_dict[key]=val
    #        self.group_value_dicts[group]=value_dict

    #def reset(self):
    #    for group in self.group_value_dicts:
    #        for key in self.group_value_dicts[group]:
    #            #print(key)
    #            val=self.group_value_dicts[group][key]
    #            t = type(val)
    #            if t == np.ndarray or t == list:
    #                setattr(group, key, val.copy())
    #            else:
    #                setattr(group, key, val)
    #
    #    for timestep in self.behaviour_timesteps:
    #        for ng in self.NeuronGroups:
    #            if timestep in ng.behaviour:
    #                ng.behaviour[timestep].reset()

    #def get_all_params(self):
    #    param_lists={}
    #    for ng in self.NeuronGroups:
    #        param_lists[ng] = [key for key in ng.__dict__]
    #    for sg in self.SynapseGroups:
    #        param_lists[sg] = [key for key in sg.__dict__]
    #    return param_lists

    #def Edge_Connect_Synapse_Group(self):
    #    #todo
    #    return


    #def split_connect(src,src_group, dst_group, x_blocks, y_blocks, receptive_field, network, syn_type):
    #    if x_blocks > 1 or y_blocks > 1:
    #        min_x = np.min(dst_group.x)
    #        max_x = np.max(dst_group.x) + 0.0001
    #        x_steps = np.arange(x_blocks + 1) * ((max_x - min_x) / x_blocks) + min_x

    #        min_y = np.min(dst_group.y)
    #        max_y = np.max(dst_group.y) + 0.0001
    #        y_steps = np.arange(y_blocks + 1) * ((max_y - min_y) / y_blocks) + min_y

            # print(x_steps,y_steps)

            # block_mat = np.empty(shape=(y_blocks, x_blocks), dtype=object)
    #        for y in range(y_blocks):
    #            for x in range(x_blocks):
    #                dst = dst_group.subGroup(
    #                    (dst_group.x >= x_steps[x]) * (dst_group.x < x_steps[x + 1]) * (dst_group.y >= y_steps[y]) * (
    #                                dst_group.y < y_steps[y + 1]))
    #                src = src_group.subGroup((src_group.x >= (x_steps[x] - receptive_field)) * (
    #                            src_group.x < (x_steps[x + 1] + receptive_field)) * (
    #                                                     src_group.y >= (y_steps[y] - receptive_field)) * (
    #                                                     src_group.y < (y_steps[y + 1] + receptive_field)))

    #                sg = SynapseGroup(src, dst, receptive_field, net=network).add_tag(syn_type)
    #    else:
    #        sg = SynapseGroup(src_group, dst_group, receptive_field, net=None).add_tag(syn_type)
    #        network.SynapseGroups.append(sg)




from PymoNNto.NetworkBehaviour.Logic.Basics.BasicHomeostasis import *

class STDP_base(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('STDP')
        #neurons.STDP_Function = self.get_init_attr('STDP_Function', [[1, 1.0]])
        #neurons.zerotime = self.get_init_attr('zerotime', 0)

        self.post_learn_value = self.get_init_attr('post_learn_value', 0.001, neurons)#0.003 better
        self.exponent = self.get_init_attr('exponent', 4.0, neurons)

        for s in neurons.afferent_synapses['GLU']:
            s.weight_change = s.get_synapse_mat()

        neurons.weight_change_sum=0#todo remove


    def new_iteration(self, neurons):
        if neurons.learning:
            weight_change_sum=0#todo remove
            for s in neurons.afferent_synapses['GLU']:
                s.weight_change *= 0

                #act_sum = s.get_dest_vec_obj(s.dst.output_activity_history[self.zerotime])[:]#output:auch lernen filtern / input:output filtern aber lernen nicht
                act_sum = s.dst.output_activity_history[self.zerotime][:]
                self.reward = np.clip(np.power(act_sum, self.exponent), 0, 1) * self.post_learn_value

                for STDP_t, STDP_y in self.STDP_Function:
                    pattern_at_t = s.src.output_activity_history[self.zerotime + STDP_t][:]
                    #pattern_at_t = s.get_src_vec_obj(s.src.output_activity_history[self.zerotime+STDP_t])[:]

                    s.weight_change += np.outer(self.reward * STDP_y, pattern_at_t)# * s.W_enabled

                s.weight_change *= s.enabled
                #weight_change_sum += np.sum(s.weight_change, axis=1)#todo remove

            #neurons.weight_change_sum = (neurons.weight_change_sum*100+weight_change_sum)/101#todo remove
            #print(neurons.weight_change_sum)

    def get_shared_variable(self, name):
        if name == 'buffer_size':
            return len(self.STDP_Function)+1

class STDP_simple(STDP_base):
    STDP_Function = [[1, 1.0]]
    zerotime=0#??? -1

class STDP_medium(STDP_base):
    STDP_Function = [[-1, -0.1], [0, 0.2], [1, 1.0]]
    zerotime=1

class STDP_complex(STDP_base):
    STDP_Function = [[-2, -0.1], [-1, -0.1], [0, 0.2], [1, 1.0], [2, 0.2]]
    zerotime=2


class TemporalWeightCache(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Weight_Cache')

        for s in neurons.afferent_synapses['GLU']:
            if not hasattr(s, 'weight_change'):
                s.weight_change = s.get_synapse_mat()

        self.decay = self.get_init_attr('decay', 1, neurons)
        self.strength = self.get_init_attr('strength', 1, neurons)

        if not hasattr(neurons, 'normalize'):
            neurons.normalize = np.array([self.get_init_attr('normalize', 1, neurons)])
        else:
            neurons.normalize = np.concatenate((neurons.normalize, np.array([self.get_init_attr('normalize', 1, neurons)])))

        self.GLU_density = self.get_init_attr('GLU_density', 1.0, neurons)
        self.GLU_random_factor = self.get_init_attr('GLU_random_factor', 1.0, neurons)
        self.GLU_equal_factor = self.get_init_attr('GLU_equal_factor', 0.0, neurons)
        all_neurons_same = self.get_init_attr('all_neurons_same', False, neurons)
        self.set_weights = self.get_init_attr('set_weights', False, neurons)

        for s in neurons.afferent_synapses['GLU']:

            if self.set_weights:
                #self.initialize_synapse_attr('W', self.GLU_density, 0.0, self.GLU_strength, neurons, 'GLU', all_neurons_same)
                s.enabled *= s.get_random_synapse_mat(self.GLU_density, all_neurons_same) > 0.0
                new_syn = s.enabled * (s.get_random_synapse_mat(1.0, all_neurons_same) * self.GLU_random_factor + self.GLU_equal_factor)
                #print(new_syn)
            else:
                new_syn = s.get_synapse_mat()#s.W.copy()

            if not hasattr(s, 'W_Caches'):
                self.cache_number = 0
                s.W_Caches = np.array([new_syn])
            else:
                self.cache_number = len(s.W_Caches)
                s.W_Caches = np.concatenate((s.W_Caches, np.array([new_syn])))


    def new_iteration(self, neurons):
        if neurons.learning:
            for s in neurons.afferent_synapses['GLU']:
                s.W_Caches[self.cache_number] = s.enabled*np.clip(s.W_Caches[self.cache_number]+s.weight_change*self.strength, 0, None)
                if self.decay != 0:
                    s.W_Caches[self.cache_number] *= self.decay

    def clear_weights(self, neurons):
        for s in neurons.afferent_synapses['GLU']:
            s.W_Caches[self.cache_number] *= 0


class RandomWeightFluctuation2(TemporalWeightCache):
    def set_variables(self, neurons):
        super().set_variables(neurons)
        for s in neurons.afferent_synapses['GLU']:
            s.W_Caches[self.cache_number]*=0
        #neurons.alpha = self.get_init_attr('alpha', 1)
        self.fluctuation = self.get_init_attr('fluctuation', 0.001, neurons)
        self.density = self.get_init_attr('density', 0.001, neurons)


    def new_iteration(self, neurons):
        if neurons.learning:
            for s in neurons.afferent_synapses['GLU']:
                s.W_Caches[self.cache_number] *= 0
                s.W_Caches[self.cache_number] += (s.get_random_synapse_mat(density=self.density))*self.fluctuation #only positive weight change
                #print(s.W_Caches[self.cache_number])
        super().new_iteration(neurons)


class DopamineProcessing(Behaviour):

    def set_variables(self, neurons):
        self.source_weight_id = self.get_init_attr('source_weight_id', 1, neurons)
        self.target_weight_id = self.get_init_attr('target_weight_id', 0, neurons)
        neurons.dopamine_level = 0

    def new_iteration(self, neurons):
        if neurons.learning:
            for s in neurons.afferent_synapses['GLU']:
                change = s.W_Caches[self.source_weight_id]*np.clip(neurons.dopamine_level, 0, 1)
                s.W_Caches[self.target_weight_id] += change
                s.W_Caches[self.source_weight_id] -= change

                #print(neurons.dopamine_level, np.sum(change))

            neurons.dopamine_level = 0

class RandomWeightFluctuation(TemporalWeightCache):
    def set_variables(self, neurons):
        super().set_variables(neurons)
        #neurons.alpha = self.get_init_attr('alpha', 1)
        self.beta = self.get_init_attr('beta', 1, neurons)
        self.gamma = self.get_init_attr('gamma', 1, neurons)


    def new_iteration(self, neurons):
        if neurons.learning:

            for s in neurons.afferent_synapses['GLU']:
                w = s.W.shape[0]
                h = s.W.shape[1]

                stable_weight=(s.W-s.W_Caches[self.cache_number])

                alpha_rnd = np.random.rand(w, h) * stable_weight * (self.beta - self.gamma)
                beta_rnd = np.random.rand(w, h) * self.beta
                gamma_rnd = np.random.rand(w, h) * self.gamma

                new_weight = s.W_Caches[self.cache_number] + (-beta_rnd+gamma_rnd) * s.W + alpha_rnd #+weight change?

                s.W_Caches[self.cache_number] = s.enabled*new_weight#np.clip(new_weight, 0, None)





from PymoNNto.NetworkCore.Behaviour import *
import numpy as np
from sympy import symbols
from sympy.physics.units import *


class ClockModule(Behaviour):

    def set_variables(self, neuron_or_network):
        self.add_tag('Clock')

        #convert_to(seconds / tau, seconds)

        neuron_or_network.clock_step_size = float(convert_to(eval(self.get_init_attr('step', '1*ms')), seconds)/second) #in ms (*ms)
        self.clock_step_size=neuron_or_network.clock_step_size
        print(neuron_or_network.clock_step_size)
        neuron_or_network.def_dtype = 0.0

    def new_iteration(self, neuron_or_network):
        neuron_or_network.def_dtype += neuron_or_network.clock_step_size

    def time_to_iterations(self, time_str):
        iterations = int(convert_to(eval(time_str + '/seconds'), seconds)/self.clock_step_size)
        print(iterations)
        return iterations



from PymoNNto.NetworkBehaviour.Input.Pattern_Basics import *
import random
import numpy as np
from collections import Counter
import random

#{' PERIOD': '.', ' COMMA': '.', ' QUESTION': '.'} #, 'MNAME':'Name', 'FNAME':'Name'
#CHILDES_activator = TextActivator_New(filename='CHILDES.txt', replace_dict={' PERIOD': '.', ' COMMA': '.', ' QUESTION': '.'})
#SPD_activator = TextActivator_New(filename='SPD.txt', replace_dict={'?': '.', ':': '.', ',': '.', '.': '. ', '  ': ' ', '-': ''})


class TextActivator_New(PatternGroup):

    def plot_char_frequency_histogram(self, bins=20):
        import matplotlib.pyplot as plt
        char_count = {}
        for s in self.corpus_blocks:
            for c in s:
                char_count[c] = 0
        sum_car = 0
        for s in self.corpus_blocks:
            for c in s:
                char_count[c] += 1
                sum_car += 1

        for c in char_count:
            char_count[c] /= sum_car

        print(char_count)

        plt.hist(list(char_count.values()), bins=bins)
        plt.show()



    def get_unique_block_text(self):
        return ' '.join(self.corpus_blocks)

    def build_sentences_recurrent(self, string, blocks, result):
        if len(blocks) > 0:
            words = blocks.pop(0)
            if len(words) == 0:
                self.build_sentences_recurrent(string, blocks.copy(), result)
            else:
                for word in words:
                    if word == '':
                        self.build_sentences_recurrent(string, blocks.copy(), result)
                    else:
                        self.build_sentences_recurrent(string + ' ' + word, blocks.copy(), result)
        else:
            return result.append(string)

    def get_char_input_statistics_dict(self):
        result = {}
        for sentence in self.corpus_blocks:
            for c in sentence:
                result[c] += 1
        return result

    def get_char_input_statistics_list(self):
        result = np.zeros(self.get_alphabet_length())
        for sentence in self.corpus_blocks:
            for c in sentence:
                result[self.char_to_index(c)] += 1
        return result

    def plot_char_input_statistics(self):
        import matplotlib.pyplot as plt

        result = self.get_char_input_statistics_list()

        objects = list(self.alphabet)
        y_pos = np.arange(len(objects))

        plt.barh(y_pos, result, align='center', alpha=0.5)
        plt.yticks(y_pos, objects)
        plt.show()
        return self

    #def get_activation_list(self):
    #    return dict([(char, source.get_activation(i, Network_UI.network['prediction_source', 0])) for i, char in enumerate(source.alphabet)])

    def build_sentences_from_sentence_bilding_blocks(self, blocks):
        result=[]
        for sentence in blocks:
            self.build_sentences_recurrent('', sentence, result)
        return result

    def make_string_redundant(self, str, redundancy_number):
        new_str = ''
        for c in str:
            for _ in range(redundancy_number):
                new_str += c
        return new_str

    def add_redundancy(self, redundancy_number):
        self.corpus = self.make_string_redundant(self.corpus, redundancy_number)

        for i, cb in enumerate(self.corpus_blocks):
            self.corpus_blocks[i] = self.make_string_redundant(cb, redundancy_number)


    def initialize(self):
        self.last_iteration=-1
        self.nextchar=None

        self.max_iterations = self.kwargs.get('max_iterations', None)
        self.random_blocks = self.kwargs.get('random_blocks', True)

        self.corpus = self.get_input_string()
        self.corpus = self.corpus.lower()

        self.corpus_blocks = self.get_text_blocks()

        #self.add_redundancy(3)

        ut = self.get_unique_block_text()
        self.alphabet = ''.join(sorted(set(ut)))
        self.words = list(sorted(set(self.split_text_into_words(ut))))

        self.current_index = -1
        self.blockcount = -1
        #self.W = self.generate_connection_e()
        #self.Input_Weight_Dict={}
        #self.input_mask=(np.sum(self.W, axis=1)>0)

        if self.kwargs.get('random_blocks', True):
            self.corpus = []

    def load_from_file(self, filename, replace_dict={}):
        data = open(filename, "r").read()
        data = data.replace('\r\n', '').replace('\r', '').replace('\n', '')
        for key in replace_dict:
            data = data.replace(key, replace_dict[key])
        return data

    def get_input_string(self):#overrride
        filename = self.kwargs.get('filename', None)
        replace_dict = self.kwargs.get('replace_dict', {})
        if filename is not None:
            return self.load_from_file(filename, replace_dict)[0:self.max_iterations]
        else:
            return self.kwargs.get('input_string', '')[0:self.max_iterations]

    def get_text_blocks(self):#override
        return self.split_text_into_sentences(self.corpus)#np.unique()#list(sorted(set()))

    def split_text_into_sentences(self, text):
        return [sentence + '.' for sentence in text.split('.')]

    def split_text_into_words(self, sentence):
        return [word for word in sentence.replace('.', ' ').replace(',', ' ').replace('?', ' ').replace('!', ' ').split(' ') if word!='']


    def generate_weights(self, neurons):

        if hasattr(neurons, 'input_density'):
            density = neurons.input_density
        else:
            density = self.kwargs.get('input_density', 1/60)#int(output_size / 60)

        output_size = neurons.size

        #output_size = self.kwargs.get('output_size', 600)
        # self.kwargs.get('activation_size', int(output_size / 60))

        if density<=1:
            self.activation_size = int(output_size * density)
        else:
            self.activation_size = int(density)

        neurons.Input_Weights = np.zeros((output_size, self.get_alphabet_length()))
        available = set(range(output_size))

        frequency_adjustment = self.kwargs.get('frequency_adjustment', False)
        char_count = self.get_char_input_statistics_list()
        mean_char_count = np.mean(char_count)
        sum_char_count = np.sum(char_count)

        for a in range(self.get_alphabet_length()):

            char_activiation_size=self.activation_size

            if frequency_adjustment:
                char_activiation_size = int(self.activation_size*(char_count[a]/mean_char_count))

                avg_char_act = char_count[a]/sum_char_count
                avg_cluster_red = char_activiation_size/28
                avg_cluster_act = avg_cluster_red * 0.02
                print(self.index_to_char(a), ': ', char_activiation_size, char_count[a], np.round(avg_char_act,decimals=4), np.round(avg_cluster_red,decimals=4), np.round(avg_cluster_act,decimals=4))

            temp = random.sample(available, char_activiation_size)
            neurons.Input_Weights[temp, a] = 1
            available = set([n for n in available if n not in temp])
            assert len(available) > 0, 'Input alphabet too big for non-overlapping neurons'

        neurons.Input_Mask = (np.sum(neurons.Input_Weights, axis=1) > 0)
        neurons.mean_network_input_activity=self.activation_size/output_size
        neurons.add_tag('text_input_group')

    def index_to_char(self, index):
        return self.alphabet[index]

    def char_to_index(self, char):
        return self.alphabet.find(char)

    def get_next_block(self):
        self.blockcount += 1
        if self.blockcount == len(self.corpus_blocks):
            self.blockcount = 0

        if self.random_blocks:
            return random.choice(self.corpus_blocks)
        else:
            return self.corpus_blocks[self.blockcount]

    def set_next_char(self, char):
        self.nextchar=char

    def get_char(self, next=False):

        if next:
            self.current_index += 1

        if self.max_iterations is not None and self.current_index % self.max_iterations == 0:
            self.current_index = 0

        if self.current_index >= len(self.corpus):
            self.corpus += self.get_next_block()

        if self.nextchar is not None:#outside activation
            self.corpus[self.current_index] = self.nextchar
            self.nextchar=None

        return self.corpus[self.current_index]

    def get_alphabet_length(self):
        return len(self.alphabet)

    def initialize_neuron_group(self, neurons):
        #if not hasattr(neurons, 'Input_Weights'):
        self.generate_weights(neurons)

    def get_activation(self, char_indx, neurons):
        return neurons.Input_Weights[:, char_indx]

    def get_pattern(self, neurons):
        if self.active:
            c = self.get_char(next=neurons.iteration != self.last_iteration)
            self.last_iteration=neurons.iteration
            i = self.char_to_index(c)
            self.current_pattern_index = i #for pattern activator
            return self.get_activation(i, neurons) #self.W[:, i]
        else:
            return None

    def add_to_dict(self, d, key):
        if key not in d:
            d[key] = 1
        else:
            d[key] += 1

    def get_all_grammar_transitions(self):
        transitions = {'. ':len(self.corpus_blocks)}

        for sentence in self.corpus_blocks:
            sl = len(sentence)
            for start in range(sl):
                for end in range(start+2, sl+1):
                    self.add_to_dict(transitions, sentence[start:end])

        return list(transitions.keys())


    def get_text_score(self, spont_output):
        result = {}
        result['output_text'] = spont_output

        #store char transitions
        result['n_different_transitions'] = 0
        for i in range(len(spont_output) - 1):
            if spont_output[i] != spont_output[i + 1]:
                result['n_different_transitions'] += 1

        if (len(spont_output) - 1)>0:
            result['%_different_transitions'] = result['n_different_transitions']/(len(spont_output) - 1)
        else:
            result['%_different_transitions'] = 0.0


        result['transition_score'] = 0
        transitions = self.get_all_grammar_transitions()
        lso = len(spont_output)
        for transition in transitions:
            found = spont_output.count(transition)
            all_ct = lso-len(transition)
            if all_ct>0:
                result['transition_score'] += np.sqrt(found)/all_ct/10

        #words and sentences
        output_sentences = self.split_text_into_sentences(spont_output)
        if len(output_sentences) > 2:
            output_sentences = output_sentences[1:-1]

        result['n_output_sentences'] = len(output_sentences)
        result['n_wrong_sentences'] = 0
        result['n_new_sentences'] = 0
        result['n_right_sentences'] = 0
        result['n_words'] = 0
        result['n_wrong_words'] = 0

        result['right_words'] = {}
        result['right_sentences'] = {}
        result['new_sentences'] = {}
        result['unique_sentences'] = {}
        result['wrong_sentences'] = {}

        for sentence in output_sentences:

            self.add_to_dict(result['unique_sentences'], sentence)

            words_in_sentence = self.split_text_into_words(sentence)
            result['n_words'] += len(words_in_sentence)

            wrong = False
            for word in words_in_sentence:
                if word not in self.words:
                    wrong = True
                    result['n_wrong_words'] += 1
                else:
                    self.add_to_dict(result['right_words'], word)

            if sentence not in self.corpus_blocks:
                if not wrong:
                    result['n_new_sentences'] += 1
                    self.add_to_dict(result['new_sentences'], sentence)
                else:
                    result['n_wrong_sentences'] += 1
                    self.add_to_dict(result['wrong_sentences'], sentence)
            else:
                result['n_right_sentences'] += 1
                self.add_to_dict(result['right_sentences'], sentence)

        result['n_right_words'] = result['n_words'] - result['n_wrong_words']

        if result['n_words'] > 0:
            result['%_right_words'] = result['n_right_words'] / result['n_words']
            result['%_wrong_words'] = result['n_wrong_words'] / result['n_words']
        else:
            result['%_right_words'] = 0.0
            result['%_wrong_words'] = 1.0

        def square_score(count_list):
            count_list=np.array(count_list)
            np.sum(np.sqrt(w_counts)) / len(self.words) / 10

        w_counts = np.array(list(result['right_words'].values()))
        result['right_word_square_score'] = np.sum(np.sqrt(w_counts))/len(self.words)/10

        rs_counts = np.array(list(result['right_sentences'].values()))
        result['right_sentences_square_score'] = np.sum(np.sqrt(rs_counts))/len(self.corpus_blocks)/10

        ns_count = np.array(list(result['new_sentences'].values()))
        result['new_sentences_square_score'] = np.sum(np.sqrt(ns_count))/len(self.corpus_blocks)/10

        result['total_score'] = (result['right_word_square_score'] + result['right_sentences_square_score'] + result['transition_score'])/3 # + result['%_different_transitions']

        #with open('obj/'+ name + '.pkl', 'wb') as f:
        #    pickle.dump(result, f, pickle.HIGHEST_PROTOCOL)

        return result


    def print_test(self, length=1000):
        temp = self.current_index
        t = ''
        for i in range(length):
            t += self.get_char(True)
        print(t)
        self.current_index = temp
        return self

    def mark_with_grammar(self, str, html=True):

        output_sentences = self.split_text_into_sentences(str)

        new_sentences = []

        for sentence in output_sentences:

            words_in_sentence = self.split_text_into_words(sentence)

            wrong = False
            for word in words_in_sentence:
                if word not in self.words:
                    wrong = True

            if sentence not in self.corpus_blocks and ' '+sentence not in self.corpus_blocks:
                if not wrong and len(words_in_sentence) >= 3:
                    if sentence not in new_sentences:
                        new_sentences.append(sentence)

        #print(new_sentences)

        mark_dict={}

        #random_color=['0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F']

        #for i, s in enumerate(self.corpus_blocks):
        #    rnd_add = random_color[i]+'0'
        #    mark_dict["background-color:#00FF"+rnd_add] = [s]

        mark_dict["background-color:#00FF00"] = self.corpus_blocks #right
        mark_dict["background-color:#FFFF00"] = new_sentences #new
        mark_dict["color:#0000FF"] = [' ' + w for w in self.words] #right words


        print(mark_dict)


        return self.mark_text(str, mark_dict, html)

    def mark_text(self, str, styles, html=True):  # styles={"background-color:#FF0000":[text1,text2,...], }
        result = str
        tags = ['a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z']
        i=0
        for style, mark_str_list in styles.items():

            for mark_str in mark_str_list:
                result = result.replace(mark_str, '<'+tags[i]+' style="' + style + '";>' + mark_str + '</'+tags[i]+'>')

            i+=1

        if html:
            return """<html>
            <head></head>
            <body>""" + result + """</body>
            </html>"""
        else:
            return result

class FDTGrammarActivator_New(TextActivator_New):

    def get_text_blocks(self):

        self.steps_plastic=self.kwargs.get('steps_plastic', 1000)
        n_removed_sentences=self.kwargs.get('n_removed_sentences', 8)


        self.verbs = ['eats ',
                      'drinks ']

        self.subjects = [' man ',
                         ' woman ',
                         ' girl ',
                         ' boy ',
                         ' child ',
                         ' cat ',
                         ' dog ',
                         ' fox ']

        self.objects_eat = ['meat.',
                            'bread.',
                            'fish.',
                            'vegetables.']

        self.objects_drink = ['milk.',
                              'water.',
                              'juice.',
                              'tea.']

        self.objects = [self.objects_eat, self.objects_drink]

        # create a huge list with all input sentences
        partial_input_string = []
        for _ in range(self.steps_plastic):
            # create a lot of sentences!
            sub = random.choice(self.subjects)
            ver = random.choice(self.verbs)
            if ver == self.verbs[0]:
                obj = random.choice(self.objects_eat)
            elif ver == self.verbs[1]:
                obj = random.choice(self.objects_drink)
            partial_input_string.append((sub + ver + obj).lower())

        # all unique input sentences
        self.all_sentences = np.unique(partial_input_string)

        # remove random sentences
        shuffled_unique_sentences = np.unique(partial_input_string)
        np.random.shuffle(shuffled_unique_sentences)

        # TODO: make sure all words appear at least once
        # TODO: sentences must be removed at random, but this was done in
        # the original work
        self.removed_sentences = []
        if n_removed_sentences >= 8:
            self.removed_sentences.extend([' woman drinks milk.',
                                           ' fox drinks tea.',
                                           ' cat eats vegetables.',
                                           ' girl eats meat.',
                                           ' child eats fish.',
                                           ' boy drinks juice.',
                                           ' man drinks water.',
                                           ' dog eats bread.'])
        if n_removed_sentences >= 16:
            self.removed_sentences.extend([' woman eats meat.',
                                           ' fox eats bread.',
                                           ' cat drinks tea.',
                                           ' girl drinks juice.',
                                           ' child drinks water.',
                                           ' boy eats fish.',
                                           ' man eats vegetables.',
                                           ' dog drinks milk.'])
        if n_removed_sentences >= 24:
            self.removed_sentences.extend([' woman eats vegetables.',
                                           ' fox eats fish.',
                                           ' cat drinks juice.',
                                           ' girl drinks tea.',
                                           ' child drinks milk.',
                                           ' boy eats bread.',
                                           ' man eats meat.',
                                           ' dog drinks water.'])
        if n_removed_sentences >= 32:
            self.removed_sentences.extend([' woman drinks water.',
                                           ' fox drinks juice.',
                                           ' cat eats bread.',
                                           ' girl eats fish.',
                                           ' child eats vegetables.',
                                           ' boy drinks tea.',
                                           ' man drinks milk.',
                                           ' dog eats meat.'])
        if n_removed_sentences >= 40:
            self.removed_sentences.extend([' woman drinks tea.',
                                           ' fox drinks milk.',
                                           ' cat eats meat.',
                                           ' girl eats vegetables.',
                                           ' child eats bread.',
                                           ' boy drinks water.',
                                           ' man drinks juice.',
                                           ' dog eats fish.'])
        if n_removed_sentences >= 48:
            self.removed_sentences.extend([' woman eats fish.',
                                           ' fox eats meat.',
                                           ' cat drinks milk.',
                                           ' girl drinks water.',
                                           ' child drinks juice.',
                                           ' boy eats vegetables.',
                                           ' man eats bread.',
                                           ' dog drinks tea.'])
        if n_removed_sentences >= 56:
            self.removed_sentences.extend([' woman drinks juice.',
                                           ' fox drinks water.',
                                           ' cat eats fish.',
                                           ' girl eats bread.',
                                           ' child eats meat.',
                                           ' boy drinks milk.',
                                           ' man drinks tea.',
                                           ' dog eats vegetables.'])

        input_string = [x for x in partial_input_string if x not in self.removed_sentences]
        #self.used_sentences = np.unique(input_string)
        return list(sorted(set(input_string)))

    def get_text_score_old(self, spont_output):
        result = {}
        output_sentences = self.split_text_into_sentences(spont_output)
        if len(output_sentences)>2:
            output_sentences=output_sentences[1:-1]
        # new output sentences
        new_dict = Counter([s for s in output_sentences if s in self.removed_sentences])
        result['n_new'] = sum(new_dict.values())
        # wrong output sentences
        wrong_dict = Counter([s for s in output_sentences if s not in self.all_sentences])
        result['n_wrong'] = sum(wrong_dict.values())
        return result

class FewLongSentencesGrammar(TextActivator_New):

    def get_text_blocks(self):
        return [' fox is an animal that eats meat.', ' boy likes to drink juice in the summer.']#, ' whale.', ' bird.', 'dog.', ' penguin likes cold weather and ice.'

class SingleWordGrammar(TextActivator_New):

    def get_text_blocks(self):
        return [' fox.', ' boy.', ' penguin.']#, ' whale.', ' bird.', 'dog.'

class FewSentencesGrammar(TextActivator_New):

    def get_text_blocks(self):
        return [' fox eats meat.', ' boy drinks juice.', ' penguin likes ice.', ' deer lives in forest.', ' parrots can fly.']#' the fish swims.' , ' deer lives in the forest.',  , ]#, ' penguin.' #,  , ' the fish swims.' #, 'the fish swims.'

class FewSentencesContextGrammar(TextActivator_New):

    def get_text_blocks(self):
        return [' fox eats meat.', ' boy eats bread.', ' penguin eats fish.', ' parrot eats nuts.']## , ]#, ' penguin.' #,  , ' the fish swims.' #

class NewGrammar(TextActivator_New):

    def get_text_blocks(self):
        return [' fox eats meat.', ' man drinks juice.', ' penguin likes ice.', ' parrots can fly.', 'the fish swims']

class Combined_text_analysis_grammar(TextActivator_New):

    def get_text_blocks(self):
        return [' fox is an animal that eats meat.', ' boy likes to drink juice in the summer.', ' fox eats meat.', ' boy drinks juice.', ' man drinks juice.', ' penguin likes ice.', ' deer lives in forest.', ' parrots can fly.', 'the fish swims', ' parrots can fly.', ' boy eats bread.', ' penguin eats fish.', ' parrot eats nuts.']


class LongDelayGrammar(TextActivator_New):

    def get_text_blocks(self):
        modes = self.kwargs.get('mode', ['simple'])#'a/the', 'where'

        if type(modes) is not list:
            modes = [modes]

        self.valid_sentence_blocks = []

        if 'a/the' in modes:
            start = ['a', 'the']
        else:
            start = []

        subject_forest = ['deer', 'fox', 'wolf', 'reindeer']
        subject_air = ['bird', 'raven', 'parrot']
        subject_water = ['fish', 'whale', 'dolphin']
        subject_ice = ['penguin', 'polar bear']

        if 'very simple' in modes or 'likes' in modes:
            center = ['likes']

            forest = ['forest.']
            air = ['trees.']
            water = ['sea.']
            ice = ['the ice.']

            self.valid_sentence_blocks.append([start, subject_forest, center, forest])
            self.valid_sentence_blocks.append([start, subject_air, center, air])
            self.valid_sentence_blocks.append([start, subject_water, center, water])
            self.valid_sentence_blocks.append([start, subject_ice, center, ice])

        if 'simple' in modes:
            center = ['likes', 'lives in', 'prefers']

            forest = ['forest.']
            air = ['trees.']
            water = ['sea.']
            ice = ['the ice.']

            self.valid_sentence_blocks.append([start, subject_forest, center, forest])
            self.valid_sentence_blocks.append([start, subject_air, center, air])
            self.valid_sentence_blocks.append([start, subject_water, center, water])
            self.valid_sentence_blocks.append([start, subject_ice, center, ice])

        if 'level31' in modes or 'where' in modes:
            where_center = ['is an animal that lives in the']

            where_forest = ['forest.']
            where_air = ['trees.']
            where_water = ['sea.']
            where_ice = ['ice.']

            self.valid_sentence_blocks.append([start, subject_forest, where_center, where_forest])
            self.valid_sentence_blocks.append([start, subject_air, where_center, where_air])
            self.valid_sentence_blocks.append([start, subject_water, where_center, where_water])
            self.valid_sentence_blocks.append([start, subject_ice, where_center, where_ice])

        if 'level17' in modes or 'what' in modes:
            what_center = ['is an animal that']

            what_forest = ['is well camouflaged.']
            what_air = ['flies to the air.']
            what_water = ['swims to the sea.']
            what_ice = ['likes cold weather.']

            self.valid_sentence_blocks.append([start, subject_forest, what_center, what_forest])
            self.valid_sentence_blocks.append([start, subject_air, what_center, what_air])
            self.valid_sentence_blocks.append([start, subject_water, what_center, what_water])
            self.valid_sentence_blocks.append([start, subject_ice, what_center, what_ice])

        if 'level17_short' in modes or 'what_short' in modes:
            what_center = ['is an animal that']

            what_forest = ['camouflages.']
            what_air = ['flies.']
            what_water = ['swims.']
            what_ice = ['likes cold weather.']

            self.valid_sentence_blocks.append([start, subject_forest, what_center, what_forest])
            self.valid_sentence_blocks.append([start, subject_air, what_center, what_air])
            self.valid_sentence_blocks.append([start, subject_water, what_center, what_water])
            self.valid_sentence_blocks.append([start, subject_ice, what_center, what_ice])

        all_sentences = self.build_sentences_from_sentence_bilding_blocks(self.valid_sentence_blocks)
        return all_sentences

class ComplexGrammarActivator(TextActivator_New):

    def get_text_blocks(self):

        the = ['a',
               'the',
               'a young',
               'the young',
               'some',
               '']

        human = ['man',
                 'woman',
                 'girl',
                 'boy',
                 'child']

        animal = ['cat',
                  'dog',
                  'fox']

        eat = ['eats']

        drink = ['drinks']

        sits = ['sits']

        objects_eat = ['meat',
                       'bread',
                       'fish',
                       'vegetables']

        objects_drink = ['milk',
                         'water',
                         'juice',
                         'tea']

        where_human = ['at the table',
                       'in the park',
                       'in a park',
                       'next to a park',
                       'in the kitchen',
                       'at the desk',
                       'on the desk']

        where_animal = ['next to forest',
                        'in the forest',
                        'in the park']

        additional = ['while the sun is shining',
                        'while its raining',
                        'and looks into the sky',
                        '']

        self.valid_sentence_blocks=[]


        self.valid_sentence_blocks.append([the, human, sits, where_human, additional])

        self.valid_sentence_blocks.append([the, animal, sits, where_human, additional])

        self.valid_sentence_blocks.append([the, human, eat, objects_eat, additional])
        self.valid_sentence_blocks.append([the, human, drink, objects_drink, additional])
        self.valid_sentence_blocks.append([the, animal, eat, objects_eat, additional])
        self.valid_sentence_blocks.append([the, animal, drink, objects_drink, additional])

        self.valid_sentence_blocks.append([the, human, eat, objects_eat, where_human, additional])
        self.valid_sentence_blocks.append([the, human, drink, objects_drink, where_human, additional])
        self.valid_sentence_blocks.append([the, animal, eat, objects_eat, where_animal, additional])
        self.valid_sentence_blocks.append([the, animal, drink, objects_drink, where_animal, additional])

        all_sentences = []

        self.build_sentences_from_sentence_bilding_blocks(self.valid_sentence_blocks, all_sentences)

        return all_sentences




class FDTGrammarActivator_New_shuffle(FDTGrammarActivator_New):

    def get_next_block(self):
        block=super().get_next_block()
        l = list(block)
        random.shuffle(l)
        result = ''.join(l)
        return result

class Random_Text_Activator(TextActivator_New):

    def initialize(self):
        self.alph = self.kwargs.get('alph', '. abcdefghijklmnopqrstuvwxyz')
        super().initialize()


    def get_random_char(self):
        rnd=np.random.randint(0, 25)
        return self.alph[rnd]

    def get_input_string(self):
        return self.alph

    def get_next_block(self):
        return self.get_random_char()

    #def get_text_blocks(self):




import numpy as np


def normalize_synapse_attr(src_attr, target_attr, target_value, neurons, synapse_type):
    neurons.temp_weight_sum = neurons.get_neuron_vec()

    for s in neurons.afferent_synapses[synapse_type]:
        s.dst.temp_weight_sum += np.sum(np.abs(getattr(s, src_attr)), axis=1)

    neurons.temp_weight_sum /= target_value

    # print(neurons.temp_weight_sum)

    for s in neurons.afferent_synapses[synapse_type]:
        setattr(s, target_attr,
                getattr(s, target_attr) / (s.dst.temp_weight_sum[:, None] + (s.dst.temp_weight_sum[:, None] == 0)))



def normalize_synapse_attr_sparse(self, src_attr, target_attr, target_value, neurons, synapse_type):

    neurons.temp_weight_sum = neurons.get_neuron_vec()

    for s in neurons.afferent_synapses[synapse_type]:
        if 'sparse' in s.tags:
            s.dst.temp_weight_sum += np.array(getattr(s, src_attr).sum(1)).flatten()
        else:
            s.dst.temp_weight_sum += np.sum(getattr(s, src_attr), axis=1)

    neurons.temp_weight_sum /= target_value

    for s in neurons.afferent_synapses[synapse_type]:
        if 'sparse' in s.tags:
            W = getattr(s, target_attr)
            W.data /= np.array(neurons.temp_weight_sum[W.indices]).reshape(W.data.shape)
        else:
            setattr(s, target_attr, getattr(s, target_attr) / (s.dst.temp_weight_sum[:, None]+(s.dst.temp_weight_sum[:, None]==0)))



from PymoNNto.Exploration.Network_UI.TabBase import *


class info_tab(TabBase):

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):

        def add_line(text, behaviour=None, neurons=None):

            element = QCheckBox()
            element.setChecked(behaviour.behaviour_enabled)
            element.setText(text)

            def click(event):
                behaviour.behaviour_enabled = element.isChecked()
                # print(behaviour.tags, element.isChecked())

            element.stateChanged.connect(click)
            element2 = QLabel(str(behaviour.init_kwargs))

            container = QVBoxLayout()  # self.main_window

            def addArg(arg):
                h_container = QHBoxLayout()
                h_container.addWidget(QLabel(arg), stretch=1)

                edit = QLineEdit(str(behaviour.init_kwargs[arg]))
                edit.arg = arg

                def change_param(event):
                    behaviour.init_kwargs[edit.arg] = edit.text()

                edit.textChanged.connect(change_param)

                h_container.addWidget(edit, stretch=1)
                h_container.addWidget(QLabel(''), stretch=5)
                container.addLayout(h_container)

            for arg in behaviour.init_kwargs:
                addArg(arg)

            def update(event):
                behaviour.set_variables(neurons)
                element2.setText(str(behaviour.init_kwargs))

            h_container = QHBoxLayout()
            btn = QPushButton('update')  # , self.main_window
            btn.clicked.connect(update)
            h_container.addWidget(btn, stretch=2)
            h_container.addWidget(QLabel(''), stretch=5)
            container.addLayout(h_container)

            frame = QFrame()
            frame.setLayout(container)
            frame.hide()

            def hide_show(event):
                if frame.isVisible():
                    frame.hide()
                else:
                    frame.show()

            if not 'structure' in behaviour.tags:
                element2.mousePressEvent = hide_show
                link_font = QFont()  # 'SansSerif', 12
                link_font.setUnderline(True)
                # palette = QPalette()
                # palette.setColor(QPalette.Text, QtCore.Qt.blue)
                # element2.setPalette(palette)
                element2.setFont(link_font)
                element2.setStyleSheet("color: rgb(0,0,200)")


            Network_UI.Add_element(element)
            Network_UI.Add_element(element2, stretch=5)
            Network_UI.Next_H_Block()
            # self.current_H_block.addLayout(container)

            Network_UI.Add_element(frame)
            Network_UI.Next_H_Block()

        Network_UI.infotabs=[]

        for group in Network_UI.network.NeuronGroups:
            infotab = Network_UI.Next_Tab('Info: ' + str(group.tags[0]))
            Network_UI.infotabs.append(infotab)

            h = 0

            caption = QLabel(str(group.tags[0]) + ':')
            caption.setFont(QFont('SansSerif', 16))
            h += 40
            Network_UI.Add_element(caption)
            Network_UI.Next_H_Block()
            for key in group.behaviour:
                behaviour = group.behaviour[key]
                main_tag = ''
                if len(behaviour.tags)>0:
                    main_tag = behaviour.tags[0]
                add_line(str(key) + ' ' + main_tag, behaviour, group)
                h += 60

            infotab.setMaximumHeight(h)


    def update(self, Network_UI):
        return


from PymoNNto.Exploration.StorageManager.StorageManager import *

#sm = StorageManager('X_1', 'X')
#sm = StorageManager('RALN_Test_2_0', 'RALN_Test_2')
#sm = StorageManager('MemoryCapacity_2', 'MemoryCapacity')#('RALN_Test_MT_Small_0', 'RALN_Test_MT_Small')

#sm = StorageManager('SORN_One_layer_all_600_0', 'SORN_One_layer_all_600')

#sm = StorageManager('SORN_Two_layer_all_600_0', 'SORN_Two_layer_all_600')
sm = StorageManager('SORN_New_Grammar_test', 'SORN_New_Grammar_test_35')#11

#plt.plot([sm.load_param('error_{}'.format(i)) for i in range(20)])
#plt.show()

weight_limit = None#1/3#None#1/3#0.5
n_biggest = 3#None

network = sm.load_obj('net')

network.NeuronGroups = [network.NeuronGroups[0]]
network.SynapseGroups = [network.SynapseGroups[0]]

source = sm.load_obj('source')

RALN = Reconstruct_Analyze_Label_Network(network)

#max_lag = np.max([ng.iteration_lag for ng in network.NeuronGroups])
#for ng in network.NeuronGroups:
#    ng.color = (0, 0, 1/max_lag*ng.iteration_lag)

groups = [network.NeuronGroups[0]]#, network.NeuronGroups[1]
RALN.label_and_group_neurons(network.NeuronGroups[0], [source.get_activation(char) for char in range(source.get_alphabet_length())], 'W', 10)

#max_vec, distribtuion, avg = RALN.get_all_buffer_length_distribution(groups)
#print(max_vec, distribtuion, avg)
#w = 1/(len(distribtuion)+10)
#for i, d in enumerate(distribtuion):
#    plt.bar(np.arange(len(d))+((w)*(i+5))-0.5, d, width=w)
#plt.plot(np.average(np.array(distribtuion), axis=0))
#plt.show()
#RALN.analyze_synapses(source, x_attribute_name='temporal_layer', y_attribute_name='class_label', weight_attribute_name='W', groups=groups, weight_limit=1/3, visualize=True)



RALN.visualize_label_and_group_neurons(x_attribute_name='temporal_layer', y_attribute_name='class_label', weight_attribute_name='W', groups=groups, weight_limit=weight_limit, n_biggest=n_biggest, source=source)
trans = RALN.get_synapse_transition_dict(source, x_attribute_name='temporal_layer', y_attribute_name='class_label', weight_attribute_name='W', groups=groups, weight_limit=weight_limit, n_biggest=n_biggest)

freq, ocd, filtered_trans = RALN.get_transition_frequencies(trans, source)
classes = RALN.get_transition_classes(trans, freq)
RALN.plot_bars(filtered_trans)
plt.xlabel('label')
plt.ylabel('synapses')
plt.show()
RALN.plot_bars(classes)
plt.xlabel('label')
plt.ylabel('synapses')
plt.show()






import numpy as np
from PymoNNto.NetworkCore.Base import *
from numpy.random import *

class Behaviour(NetworkObjectBase):
    #modificaton_reset_vars = []
    run_on_init = False

    def __init__(self, **kwargs):
        self.init_kwargs = kwargs
        self.used_attr_keys = []
        self.behaviour_enabled = self.get_init_attr('behaviour_enabled', True, None)
        super().__init__(tag=self.get_init_attr('tag', None, None))

    def run_on_neuron_init(self):
        self.run_on_neuron_init_var = True
        return self

    def diversity_string_to_vec2(self, ds, neurons):

        if 'same(' in ds and ds[-1] == ')':
            params=ds[5:-1].replace(' ', '').split(',')
            if len(params) == 2:
                #print('same', params)
                return getattr(neurons[params[0], 0], params[1])

        command = ds.split(';')
        if (not '(' in ds and not ')' in ds) and not command[0].replace('.', '').replace('+', '').replace('-', '').isdigit():
            return ds


        if len(command) == 1:
            if command[0].replace('.', '').replace('+', '').replace('-', '').isdigit():
                result = float(command[0])

            #else:
            #    result = self.get_random_nparray((neurons.size), rnd_code=command[0])
                #print(result, type(result))
                return result

        dist_cmd='uniform(low, high)'
        if len(command) > 2 and not 'plot' in command[2]:# problem: , also in command
            dist_cmd=command[2]

            if 'smooth' in command[2]:
                smoothing = float(command[2].replace('smooth', ''))
                dist_cmd = 'np.sum([uniform(low, high, size=dim) for _ in range(smoothing)], axis=0)/smoothing'.replace('smoothing', str(smoothing))


        if command[0].replace('.', '').replace('+', '').replace('-', '').isdigit():
            min_v = float(command[0])
            if '%' in command[1]:
                max_v = min_v
                if '-' in command[1]:
                    min_v -= min_v / 100 * float(command[1].replace('+', '').replace('-', '').replace('%', ''))
                if '+' in command[1]:
                    max_v += max_v / 100 * float(command[1].replace('+', '').replace('-', '').replace('%', ''))
            else:
                max_v = float(command[1])

            dist_cmd=dist_cmd.replace('low',str(min_v)).replace('high',str(max_v)).replace('min_v',str(min_v)).replace('max_v',str(max_v)).replace('min',str(min_v)).replace('max',str(max_v))
        else:
            dist_cmd=command[0]

        if 'lognormal_real_mean' in command[0]:
            parts = command[0].replace(')', '').replace(' ', '').split('(')
            arguments = parts[1].split(',')
            mean = float(arguments[0])
            sigma = float(arguments[1])
            mu = -np.power(sigma, 2) + np.log(mean)
            dist_cmd = 'lognormal({}, {})'.format(mu, sigma)

        result=self.get_random_nparray((neurons.size), rnd_code=dist_cmd)


        if 'plot' in command[-1]:
            import matplotlib.pyplot as plt
            #print(result)
            plt.hist(result, bins=30)
            plt.show()


        return result

    def diversity_string_to_vec(self, ds, neurons):

        if 'same(' in ds and ds[-1] == ')':
            params=ds[5:-1].replace(' ', '').split(',')
            if len(params) == 2:
                #print('same', params)
                return getattr(neurons[params[0], 0], params[1])

        command = ds.split(';')
        if (not '(' in ds and not ')' in ds) and not command[0].replace('.', '').replace('+', '').replace('-', '').isdigit():
            return ds


        if len(command) == 1:
            if command[0].replace('.', '').replace('+', '').replace('-', '').isdigit():
                result = float(command[0])

            #else:
            #    result = self.get_random_nparray((neurons.size), rnd_code=command[0])
                #print(result, type(result))
                return result

        dist_cmd='uniform(low, high)'
        if len(command) > 2 and not 'plot' in command[2]:# problem: , also in command
            dist_cmd=command[2]

            if 'smooth' in command[2]:
                smoothing = float(command[2].replace('smooth', ''))
                dist_cmd = 'np.sum([uniform(low, high, size=dim) for _ in range(smoothing)], axis=0)/smoothing'.replace('smoothing', str(smoothing))


        if command[0].replace('.', '').replace('+', '').replace('-', '').isdigit():
            min_v = float(command[0])
            if '%' in command[1]:
                max_v = min_v
                if '-' in command[1]:
                    min_v -= min_v / 100 * float(command[1].replace('+', '').replace('-', '').replace('%', ''))
                if '+' in command[1]:
                    max_v += max_v / 100 * float(command[1].replace('+', '').replace('-', '').replace('%', ''))
            else:
                max_v = float(command[1])

            dist_cmd=dist_cmd.replace('low',str(min_v)).replace('high',str(max_v)).replace('min_v',str(min_v)).replace('max_v',str(max_v)).replace('min',str(min_v)).replace('max',str(max_v))
        else:
            dist_cmd=command[0]

        if 'lognormal_real_mean' in command[0]:
            parts = command[0].replace(')', '').replace(' ', '').split('(')
            arguments = parts[1].split(',')
            mean = float(arguments[0])
            sigma = float(arguments[1])
            mu = -np.power(sigma, 2) + np.log(mean)
            dist_cmd = 'lognormal({}, {})'.format(mu, sigma)

        result=self.get_random_nparray((neurons.size), rnd_code=dist_cmd)


        if 'plot' in command[-1]:
            import matplotlib.pyplot as plt
            #print(result)
            plt.hist(result, bins=30)
            plt.show()


        return result

    def set_init_attrs_as_variables(self, object):
        for key in self.init_kwargs:
            setattr(object, key, self.get_init_attr(key, None, neurons=object))
            print('init', key)
            #get_init_attr

    def check_unused_attrs(self):
        for key in self.init_kwargs:
            if not key in self.used_attr_keys:
                print('Warning: "'+key+'" not used in set_variables of '+str(self)+' behaviour! Make sure that "'+key+'" is spelled correctly and get_init_attr('+key+',...) is called in set_variables. Valid attributes are:'+str(self.used_attr_keys))

    def get_init_attr(self, key, default, neurons=None, do_not_diversify=False, search_other_behaviours=False):
        self.used_attr_keys.append(key)

        result = self.init_kwargs.get(key, default)

        if key not in self.init_kwargs and neurons is not None and search_other_behaviours:
            for b in neurons.behaviours:
                if key in b.init_kwargs:
                    result = b.init_kwargs.get(key, result)

        if not do_not_diversify and type(result) is str and neurons is not None:
            result = self.diversity_string_to_vec(result, neurons)

        return result

    def set_variables(self, neurons):
        return

    def reset(self):
        return

    def new_iteration(self, neurons):
        return

    #def get_shared_variable(self, name):
    #    return None



    def initialize_synapse_attr(self, target_attr, density, equal_range, random_range, neurons, synapse_type, all_neurons_same=False):
        for s in neurons.afferent_synapses[synapse_type]:
            s.enabled *= s.get_random_synapse_mat(density, all_neurons_same) > 0.0
            setattr(s, target_attr, s.enabled*(s.get_synapse_mat()+equal_range+s.get_random_synapse_mat(1.0, all_neurons_same)*random_range))#s.shape_mat*#neurons.GABA_strength+neurons.min_GABA_strength









#def evaluate_unit(unit_string):
#    if time in unit_string

time_units = ['seconds', 'minutes', 'hours', 'milliseconds', 'ms']

seconds = 1.0
minutes = seconds*60
hours = minutes*60
milliseconds = seconds/1000
ms = milliseconds

units = time_units+[]

def initialize_object_variable_unit_dict(obj):
    if not hasattr(obj, 'variable_unit_dict'):
        obj.variable_unit_dict = {}

def split_equation_and_unit(str):
    str=str.replace(' ','')
    parts = []
    str_buf = ''
    for s in str:

        if s in ['*', '/', '+', '-', '%', ':', ';', '=', '!', '(', ')', '[', ']']:
            parts.append(str_buf)
            parts.append(s)
            str_buf = ''
        else:
            str_buf += s

    parts.append(str_buf)


    quantity_formula_parts = []
    unit_formula_parts = []
    for part in parts:
        if part in units:
            quantity_formula_parts.append('1.0')
            unit_formula_parts.append(part)
        elif not part in unit:
            quantity_formula_parts.append(part)
            unit_formula_parts.append(part)
        else:#operators
            quantity_formula_parts.append(part)
            unit_formula_parts.append(part)

    equation = ''.join(parts)
    unit = ''.join(parts)

    return equation, unit






from PymoNNto.Exploration.Network_UI.TabBase import *


class afferent_syn_attr_plot_tab(TabBase):

    def __init__(self, syn_vars, title='Aff Syn',timesteps=500):
        super().__init__(title)
        self.syn_vars=syn_vars
        self.timesteps=timesteps

    def add_recorder_variables(self, neuron_group, Network_UI):
        self.check={}
        for syn_var in self.syn_vars:
            self.check[syn_var] = True
            for syn in neuron_group.afferent_synapses["All"]:
                if not hasattr(syn, syn_var):
                    self.check[syn_var] = False
            if self.check[syn_var]:
                Network_UI.add_recording_variable(neuron_group, '[np.sum(s.'+syn_var+') for s in n.afferent_synapses["All"]]', timesteps=self.timesteps)

    def initialize(self, Network_UI):
        self.weight_tab = Network_UI.Next_Tab(self.title)

        self.plots = {}
        for syn_var in self.syn_vars:
            _, syn_plt = Network_UI.Add_plot_curve('Neuron Group average ' + syn_var, return_plot=True, x_label='t (iterations)', y_label='Input')
            self.plots[syn_var] = syn_plt



    def update(self, Network_UI):
        if self.weight_tab.isVisible() and len(Network_UI.network[Network_UI.neuron_select_group]) > 0:

            group = Network_UI.network[Network_UI.neuron_select_group, 0]

            for syn_var in self.syn_vars:

                if self.check[syn_var]:

                    self.plots[syn_var].clear()

                    recorded = group['[np.sum(s.'+syn_var+') for s in n.afferent_synapses["All"]]'][-self.timesteps:]
                    iterations = group['n.iteration', 0, 'np'][-self.timesteps:]
                    if len(recorded) > 0:
                        inputs = np.array(recorded[0])
                        ident=[s.src.group_without_subGroup() for s in group.afferent_synapses["All"]]
                        single_ident = list(set(ident))

                        for i, si in enumerate(single_ident):
                            mask = [id == si for id in ident]

                            data = np.sum(inputs[:, mask], axis=1)
                            curve = pg.PlotCurveItem(iterations, data, name='', pen=si.color)#self.slow_input_colors[i]
                            self.plots[syn_var].addItem(curve)



from PymoNNto.NetworkBehaviour.Input.Images.PixelPattern import *

class Line_Patterns(Pixel_Pattern):

    def initialize(self):
        super().initialize()

        center_x = self.kwargs.get('center_x', 1)
        center_y = self.kwargs.get('center_y', 1)
        degree = self.kwargs.get('degree', 1)
        line_length = self.kwargs.get('line_length', 1)

        pattern_count=0
        if type(center_x) in [list, np.ndarray]: pattern_count = np.maximum(pattern_count, len(center_x))
        if type(center_y) in [list, np.ndarray]: pattern_count = np.maximum(pattern_count, len(center_y))
        if type(degree) in [list, np.ndarray]: pattern_count = np.maximum(pattern_count, len(degree))
        if pattern_count == 0: pattern_count=1

        if not type(center_x) in [list, np.ndarray]: center_x=np.ones(pattern_count)*center_x
        if not type(center_y) in [list, np.ndarray]: center_y=np.ones(pattern_count)*center_y
        if not type(degree) in [list, np.ndarray]: degree=np.ones(pattern_count)*degree

        for i in range(pattern_count):
            self.patterns.append(np.array([getLinePicture(degree[i], center_x[i], center_y[i], line_length/2, self.grid_width, self.grid_height)]))
            self.labels.append('deg{}cy{}cx{}'.format(degree[i], center_x[i], center_y[i]))


from PymoNNto.NetworkBehaviour.Input.Activator import *
from PymoNNto.NetworkBehaviour.Logic.SORN.SORN_advanced_buffer import *

class SORN_init_neuron_vars_decay(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('init_neuron_vars')

        neurons.activity = neurons.get_neuron_vec()
        neurons.excitation = neurons.get_neuron_vec()
        neurons.inhibition = neurons.get_neuron_vec()
        neurons.input_act = neurons.get_neuron_vec()

        neurons.timescale = self.get_init_attr('timescale', 1)

    def new_iteration(self, neurons):
        if first_cycle_step(neurons):
            neurons.activity *= 0.9
            neurons.excitation.fill(0)# *= 0
            neurons.inhibition.fill(0)# *= 0
            neurons.input_act.fill(0)# *= 0

class WTA_refrac(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Refractory_A')
        neurons.refractory_counter_analog = neurons.get_neuron_vec()
        self.decayfactor = self.get_init_attr('decayfactor', 0.5, neurons)


    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            #neurons.activity -= neurons.refractory_counter_analog * self.strengthfactor

            neurons.refractory_counter_analog *= self.decayfactor
            neurons.refractory_counter_analog += neurons.output

class WTA_refrac_apply(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Refractory_Apply')
        self.strengthfactor = self.get_init_attr('strengthfactor', 1.0, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.activity -= neurons.refractory_counter_analog * self.strengthfactor

class SORN_slow_syn_simple(SORN_signal_propagation_base):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('slow_simple' + self.transmitter)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons) and self.strength != 0:

            for s in neurons.afferent_synapses[self.transmitter]:
                s.slow_add = s.W.dot(s.src.output) * self.strength

                s.dst.activity += s.slow_add
                if self.strength > 0:
                    s.dst.excitation += s.slow_add
                else:
                    s.dst.inhibition += s.slow_add



from PymoNNto.Exploration.Evolution.Evolution_Device_Multi_Thread import *
from PymoNNto.Exploration.StorageManager.StorageManager import *
import paramiko
from scp import SCPClient
import sys

#ausführen abhängig von pfad!!!

def get_ssh_connection(host, user, password):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(host, username=user, password=password)
    return ssh

def get_response(out, err):
    result = []
    for line in out.read().splitlines():
        result.append(line.decode("utf-8"))
    for line in err.read().splitlines():
        result.append(line.decode("utf-8"))
    return result

def ssh_thread_worker(slave_file, name, user, host, password, root_folder, conn):
    print('ssh thread started')
    while True:
        time.sleep(np.random.rand())
        if conn.poll():
            genome = conn.recv()
            try:
                print('run')
                ssh = get_ssh_connection(host, user, password)

                cmd = 'cd ' + name + '/' +root_folder+ '; '
                cmd += 'python3 ' + slave_file + ' genome=' + get_gene_id(genome)
                ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)

                results = get_response(ssh_stdout, ssh_stderr)

                ssh.close()

                #timeout
                #connection abort

                success = False
                for line in results:
                    if 'evolution score set to #' in line:
                        parts = line.split('#')
                        if len(parts) == 3:
                            genome['score'] = float(parts[1])
                            success = True


                if success:
                    sm = StorageManager(main_folder_name=genome['evo_name'], folder_name=get_gene_file(genome), print_msg=False, add_new_when_exists=False)
                    sm.save_param_dict(genome)
                    conn.send([genome, 'success'])
                else:
                    conn.send([genome, 'no score found in ssh thread results'+str(results)])


            except:
                conn.send([genome, 'thread error'])

        conn.send([None, 'idle'])
        time.sleep(1.0)




class Evolution_Device_SSH(Evolution_Device_Multi_Thread):

    def __init__(self, device_string, parent):
        super().__init__(device_string, parent)
        self.user = None
        self.host = None
        self.password = None

        temp=sys.argv[0].replace(get_data_folder().replace('Data',''), '')
        temp='/'.join(temp.split('/')[:-1])#remove file name
        self.root_folder = temp #evolution_master_path_from_project_folder

        #compute main folder dependency!!!

        user_host = self.device_string.replace('ssh ', '').split('@')
        if len(user_host) == 2:
            self.user = user_host[0]
            self.host = user_host[1]
            if ' ' in self.host:
                host_pw = self.host.split(' ')
                if len(host_pw) == 2:
                    self.host = host_pw[0]
                    self.password = host_pw[1]
                    print('warning: a clear text ssh password inside of the code is dangerous!')
                else:
                    print('space in host but no password detected')
        else:
            print('cannot split user and host')

    def initialize(self):
        self.parent_conn, child_conn = Pipe()
        self.process = Process(target=ssh_thread_worker, args=(self.parent.slave_file, self.parent.name, self.user, self.host, self.password, self.root_folder, child_conn))

    def initialize_device_group(self):

        #search main project folder
        Data_Folder=get_data_folder()
        if Data_Folder !='./Data' and False:
            Main_Folder = Data_Folder.replace('Data', '')

            src = Data_Folder + '/'+self.parent.name+'.zip'
            dst = self.parent.name+'.zip'

            # zip project
            zipDir(Main_Folder, src, ['.git', '.zip', '.idea', '\\StorageManager\\', '\\NetworkStates\\', '\\Evo\\', '\\__pycache__\\', '\\midis\\'])

            print(self.user, self.host, self.password, )

            # transfer zip
            ssh = get_ssh_connection(self.host, self.user, self.password)
            scp = SCPClient(ssh.get_transport())
            scp.put(src, dst)

            #remove zip
            os.remove(src)

            # remote unzip and remote remove zip
            cmd = 'rm -r -d ' + self.parent.name + '; '                                  #remove old directory
            cmd += 'mkdir ' + self.parent.name + '; '                                    #create new directory
            cmd += 'unzip ' + self.parent.name + '.zip -d ' + self.parent.name + '; '    #unzip
            cmd += 'rm ' + self.parent.name + '.zip'                                     #remove zip
            ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)
            ssh_stdout.channel.recv_exit_status()

            scp.close()
            ssh.close()
        else:
            print('Error No root "Data" folder found')


#ssh = get_ssh_connection('hey3kmuagjunsk2b.myfritz.net', 'marius', None)

#name = 'my_test_evo'
#root_folder = 'Exploration/Evolution'
#slave_file = 'test_slave.py'
#genome={'a':1,'b':2,'evo_name':'test','gen':1,'id':2}

#cmd = 'cd ' + name + '/' + root_folder + '; '
#cmd += 'python3 ' + slave_file + ' genome=' + get_gene_id(genome)
#ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)
#t = ssh_stdout.channel.recv_exit_status()
#ssh_stdin.close()

#for line in ssh_stdout.read().splitlines():
#    print(line.decode("utf-8"))

#for line in ssh_stderr.read().splitlines():
#    print(line.decode("utf-8"))

#time.sleep(1)

#ssh.close()

#time.sleep(1)

#print_response(ssh_stdin)
#print_response(ssh_stdout)
#print_response(ssh_stderr)


from PymoNNto.Exploration.Network_UI.TabBase import *

class sidebar_activity_sub_module(TabBase):

    def __init__(self, add_color_dict={'output': (255, 255, 255), 'Input_Mask': (-100, -100, -100)}):
        self.add_color_dict = add_color_dict
        self.compiled={}
        for param in self.add_color_dict:
            self.compiled[param] = compile('n.'+param, '<string>', 'eval')

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI, index):

        def mce(event):
            Network_UI.neuron_select_group = event.currentItem.neuron_group_tag
            w = Network_UI.network[Network_UI.neuron_select_group, 0].width
            h = Network_UI.network[Network_UI.neuron_select_group, 0].height
            d = Network_UI.network[Network_UI.neuron_select_group, 0].depth
            y_temp = int((h*d - 1) - np.trunc(event.pos().y())) #np.clip(int((h - 1) - np.trunc(event.pos().y())), 0, h - 1)

            Network_UI.neuron_select_x = np.clip(int(np.trunc(event.pos().x())), 0, w - 1)
            Network_UI.neuron_select_y = np.clip(int(y_temp-np.trunc(y_temp/h)), 0, h - 1)
            Network_UI.neuron_select_z = np.clip(int(np.trunc(y_temp/h)), 0, d - 1)

            h_abs = np.clip(int(y_temp), 0, h * d - 1)
            Network_UI.neuron_select_id = (h_abs) * w + Network_UI.neuron_select_x

            print(Network_UI.neuron_select_x, Network_UI.neuron_select_y, Network_UI.neuron_select_z, Network_UI.neuron_select_id, ' ', event.pos().x(), event.pos().y())
            Network_UI.static_update_func()

        group_select_box=QComboBox()
        Network_UI.Add_Sidebar_Element(group_select_box)

        Network_UI.group_sliders.append(QSlider(1))  # QtCore.Horizontal
        Network_UI.group_sliders[-1].setMinimum(0)
        Network_UI.group_sliders[-1].setMaximum(100)
        Network_UI.group_sliders[-1].setSliderPosition(100)
        Network_UI.group_sliders[-1].mouseReleaseEvent = Network_UI.static_update_func
        Network_UI.group_sliders[-1].setToolTip('scale neuron-group plots up and down (only visualization)')

        Network_UI.sidebar_hblock.addWidget(Network_UI.group_sliders[-1])

        self.image_item = Network_UI.Add_Image_Item(False, True, tooltip_message='white: active neurons\r\ndarker color: primary input neurons\r\ngreen: selected neuron')

        self.image_item.neuron_group_tag = Network_UI.group_tags[index]
        Network_UI.neuron_visible_groups.append(Network_UI.group_tags[index])
        self.image_item.mouseClickEvent = mce

        def group_changed(select_index):
            tag=Network_UI.group_tags[select_index]
            Network_UI.neuron_select_group = tag
            self.image_item.neuron_group_tag = tag
            Network_UI.neuron_visible_groups[index] = tag
            Network_UI.neuron_select_id = 0

        group_select_box.addItems(Network_UI.group_tags)
        group_select_box.setCurrentIndex(index)
        group_select_box.currentIndexChanged.connect(group_changed)



    def update(self, Network_UI):

        #for group_tag in Network_UI.group_tags:
        group_tag = self.image_item.neuron_group_tag
        if len(Network_UI.network[group_tag]) > 0:

            group = Network_UI.network[group_tag, 0]
            n=group #for eval operations

            alpha = group.color[3] / 255.0
            base_color = (group.color[0] * alpha, group.color[1] * alpha, group.color[2] * alpha)

            image=np.zeros((group.size,3))

            for i in range(3):
                image[:, i] += base_color[i]

            if Network_UI.neuron_select_group == group_tag:
                for i in range(3):
                    image[Network_UI.neuron_select_id, i] = Network_UI.neuron_select_color[i]

            for param, color in self.add_color_dict.items():
                try:
                #if True:#hasattr(group, param):
                    data=eval(self.compiled[param])

                    if (type(data) == np.ndarray and data.dtype == np.dtype('bool')) or (type(data) == bool):
                        for i in range(3):
                            image[data, i] += color[i]
                    else:
                        for i in range(3):
                            image[:, i] += color[i]*data
                except:
                    print(param, "can not be evaluated")

            image=np.reshape(image, (group.height*group.depth, group.width, 3))#group.depth
            self.image_item.setImage(np.rot90(image, 3), levels=(0, 255))

class UI_sidebar_activity_module():

    def __init__(self, group_display_count=1, add_color_dict={'output': (255, 255, 255), 'Input_Mask': (-100, -100, -100)}):
        self.group_display_count=group_display_count
        self.add_color_dict=add_color_dict

    def add_recorder_variables(self, neuron_group, Network_UI):
        for module in self.sub_modules:
            module.add_recorder_variables(neuron_group, Network_UI)

    def initialize(self, Network_UI):

        if Network_UI.group_display_count is not None:
            self.group_display_count=Network_UI.group_display_count

        self.sub_modules = []
        for i in range(self.group_display_count):
            self.sub_modules.append(sidebar_activity_sub_module(add_color_dict=self.add_color_dict))
            self.sub_modules[-1].initialize(Network_UI, np.minimum(i, len(Network_UI.group_tags)))

    def update(self, Network_UI):
        for module in self.sub_modules:
            module.update(Network_UI)